'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import uuid
from enum import Enum

from pm4py.objects.petri_net.utils import reduction
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.petri_net.utils.petri_utils import remove_place
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to
from pm4py.util import exec_utils, nx_utils


class Parameters(Enum):
    USE_ID = "use_id"
    ENABLE_REDUCTION = "enable_reduction"
    RETURN_FLOW_TRANS_MAP = "return_flow_trans_map"


def build_digraph_from_petri_net(net):
    """
    Builds a directed graph from a Petri net
        (for the purpose to add invisibles between inclusive gateways)

    Parameters
    -------------
    net
        Petri net

    Returns
    -------------
    digraph
        Digraph
    """
    graph = nx_utils.DiGraph()
    for place in net.places:
        graph.add_node(place.name)
    for trans in net.transitions:
        in_places = [x.source for x in list(trans.in_arcs)]
        out_places = [x.target for x in list(trans.out_arcs)]
        for pl1 in in_places:
            for pl2 in out_places:
                graph.add_edge(pl1.name, pl2.name)
    return graph


def apply(bpmn_graph, parameters=None):
    """
    Converts a BPMN graph to an accepting Petri net

    Parameters
    --------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters of the algorithm:
        - Parameters.USE_ID => (default: False) uses the IDs of the objects instead of their labels in the conversion
        - Parameters.ENABLE_REDUCTION => reduces the invisible transitions
        - Parameters.RETURN_FLOW_TRANS_MAP => returns additional information on the conversion:
                                                (iv) the places of the obtained Petri net that are corresponding to each
                                                    BPMN flow.
                                                (v) the transitions of the Petri net related to the nodes of the BPMN
                                                    diagram.

    Returns
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.bpmn.obj import BPMN

    use_id = exec_utils.get_param_value(Parameters.USE_ID, parameters, False)
    return_flow_trans_map = exec_utils.get_param_value(
        Parameters.RETURN_FLOW_TRANS_MAP, parameters, False
    )
    enable_reduction = exec_utils.get_param_value(
        Parameters.ENABLE_REDUCTION, parameters, True
    )

    if return_flow_trans_map:
        enable_reduction = False

    net = PetriNet("")
    source_place = PetriNet.Place("source")
    net.places.add(source_place)
    sink_place = PetriNet.Place("sink")
    net.places.add(sink_place)
    im = Marking()
    fm = Marking()
    im[source_place] = 1
    fm[sink_place] = 1

    # keep this correspondence for adding invisible transitions for OR-gateways
    inclusive_gateway_exit = set()
    inclusive_gateway_entry = set()

    flow_place = {}
    source_count = {}
    target_count = {}
    for flow in bpmn_graph.get_flows():
        if isinstance(flow, BPMN.SequenceFlow):
            source = flow.get_source()
            target = flow.get_target()
            place = PetriNet.Place(str(flow.get_id()))
            net.places.add(place)
            flow_place[flow] = place
            if source not in source_count:
                source_count[source] = 0
            if target not in target_count:
                target_count[target] = 0
            source_count[source] = source_count[source] + 1
            target_count[target] = target_count[target] + 1

    for flow in bpmn_graph.get_flows():
        if isinstance(flow, BPMN.SequenceFlow):
            source = flow.get_source()
            target = flow.get_target()
            place = PetriNet.Place(str(flow.get_id()))
            if (
                isinstance(source, BPMN.InclusiveGateway)
                and source_count[source] > 1
            ):
                inclusive_gateway_exit.add(place.name)
            elif (
                isinstance(target, BPMN.InclusiveGateway)
                and target_count[target] > 1
            ):
                inclusive_gateway_entry.add(place.name)

    # remove possible places that are both in inclusive_gateway_exit and inclusive_gateway_entry,
    # because we do not need to add invisibles in this situation
    incl_gat_set_inters = inclusive_gateway_entry.intersection(
        inclusive_gateway_exit
    )
    inclusive_gateway_exit = inclusive_gateway_exit.difference(
        incl_gat_set_inters
    )
    inclusive_gateway_entry = inclusive_gateway_entry.difference(
        incl_gat_set_inters
    )

    nodes_entering = {}
    nodes_exiting = {}
    trans_map = {}

    for node in bpmn_graph.get_nodes():
        if (
            isinstance(node, BPMN.Task)
            or isinstance(node, BPMN.StartEvent)
            or isinstance(node, BPMN.EndEvent)
            or isinstance(node, BPMN.ExclusiveGateway)
            or isinstance(node, BPMN.ParallelGateway)
            or isinstance(node, BPMN.InclusiveGateway)
        ):
            if node not in source_count:
                source_count[node] = 0
            if node not in target_count:
                target_count[node] = 0

            entry_place = PetriNet.Place("ent_" + str(node.get_id()))
            net.places.add(entry_place)
            exiting_place = PetriNet.Place("exi_" + str(node.get_id()))
            net.places.add(exiting_place)
            if use_id:
                label = str(node.get_id())
            else:
                label = (
                    str(node.get_name())
                    if isinstance(node, BPMN.Task)
                    else None
                )
                if not label:
                    label = None
            transition = PetriNet.Transition(
                name=str(node.get_id()), label=label
            )
            net.transitions.add(transition)
            trans_map[node] = [transition]
            add_arc_from_to(entry_place, transition, net)
            add_arc_from_to(transition, exiting_place, net)

            if isinstance(node, BPMN.ParallelGateway) or isinstance(
                node, BPMN.InclusiveGateway
            ):
                if source_count[node] > 1:
                    exiting_object = PetriNet.Transition(
                        str(uuid.uuid4()), None
                    )
                    net.transitions.add(exiting_object)
                    add_arc_from_to(exiting_place, exiting_object, net)
                    trans_map[node].append(exiting_object)
                else:
                    exiting_object = exiting_place

                if target_count[node] > 1:
                    entering_object = PetriNet.Transition(
                        str(uuid.uuid4()), None
                    )
                    net.transitions.add(entering_object)
                    add_arc_from_to(entering_object, entry_place, net)
                    trans_map[node].append(entering_object)
                else:
                    entering_object = entry_place
                nodes_entering[node] = entering_object
                nodes_exiting[node] = exiting_object
            else:
                nodes_entering[node] = entry_place
                nodes_exiting[node] = exiting_place

            if isinstance(node, BPMN.StartEvent):
                start_transition = PetriNet.Transition(str(uuid.uuid4()), None)
                net.transitions.add(start_transition)
                add_arc_from_to(source_place, start_transition, net)
                add_arc_from_to(start_transition, entry_place, net)
                trans_map[node].append(start_transition)
            elif isinstance(node, BPMN.EndEvent):
                end_transition = PetriNet.Transition(str(uuid.uuid4()), None)
                net.transitions.add(end_transition)
                add_arc_from_to(exiting_place, end_transition, net)
                add_arc_from_to(end_transition, sink_place, net)
                trans_map[node].append(end_transition)

    for flow in bpmn_graph.get_flows():
        if isinstance(flow, BPMN.SequenceFlow):
            if (
                flow.get_source() in nodes_exiting
                and flow.get_target() in nodes_entering
            ):
                source_object = nodes_exiting[flow.get_source()]
                target_object = nodes_entering[flow.get_target()]

                if isinstance(source_object, PetriNet.Place):
                    inv1 = PetriNet.Transition(f"sfl_{flow.get_id()}", None)
                    net.transitions.add(inv1)
                    add_arc_from_to(source_object, inv1, net)
                    source_object = inv1
                    trans_map[flow.source].append(inv1)

                if isinstance(target_object, PetriNet.Place):
                    inv2 = PetriNet.Transition(f"tfl_{flow.get_id()}", None)
                    net.transitions.add(inv2)
                    add_arc_from_to(inv2, target_object, net)
                    target_object = inv2
                    trans_map[flow.target].append(inv2)

                add_arc_from_to(source_object, flow_place[flow], net)
                add_arc_from_to(flow_place[flow], target_object, net)

    if inclusive_gateway_exit and inclusive_gateway_entry:
        # do the following steps if there are inclusive gateways:
        # - calculate the shortest paths
        # - add an invisible transition between couples of corresponding places
        # this ensures soundness and the correct translation of the BPMN
        inv_places = {x.name: x for x in net.places}
        digraph = build_digraph_from_petri_net(net)
        all_shortest_paths = dict(nx_utils.all_pairs_dijkstra(digraph))
        keys = list(all_shortest_paths.keys())

        for pl1 in inclusive_gateway_exit:
            if pl1 in keys:
                output_places = sorted(
                    [
                        (x, len(y))
                        for x, y in all_shortest_paths[pl1][1].items()
                        if x in inclusive_gateway_entry
                    ],
                    key=lambda x: x[1],
                )
                if output_places:
                    inv_trans = PetriNet.Transition(str(uuid.uuid4()), None)
                    net.transitions.add(inv_trans)
                    add_arc_from_to(inv_places[pl1], inv_trans, net)
                    add_arc_from_to(
                        inv_trans, inv_places[output_places[0][0]], net
                    )

    if enable_reduction:
        reduction.apply_simple_reduction(net)

    for place in list(net.places):
        if (
            len(place.in_arcs) == 0
            and len(place.out_arcs) == 0
            and place not in im
            and place not in fm
        ):
            remove_place(net, place)

    if return_flow_trans_map:
        return net, im, fm, flow_place, trans_map

    return net, im, fm



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
"""

from typing import List, Optional, Tuple, Dict, Union, Generator, Set, Any

from pm4py.objects.log.obj import Trace, EventLog, EventStream
from pm4py.utils import __event_log_deprecation_warning
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.utils import get_properties, pandas_utils, constants
from pm4py.util.pandas_utils import check_is_pandas_dataframe, check_pandas_dataframe_columns
from pm4py.util import labels_similarity as ls_util
from pm4py.util import deprecation

import pandas as pd


@deprecation.deprecated(
    deprecated_in="2.3.0",
    removed_in="3.0.0",
    details="this method will be removed in a future release.",
)
def construct_synchronous_product_net(
    trace: Trace,
    petri_net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Constructs the synchronous product net between a trace and a Petri net process model.

    :param trace: A trace from an event log.
    :param petri_net: The Petri net process model.
    :param initial_marking: The initial marking of the Petri net.
    :param final_marking: The final marking of the Petri net.
    :return: A tuple containing the synchronous Petri net, the initial marking, and the final marking.
    :rtype: Tuple[PetriNet, Marking, Marking]

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        log = pm4py.read_xes('log.xes')
        sync_net, sync_im, sync_fm = pm4py.construct_synchronous_product_net(log[0], net, im, fm)
    """
    from pm4py.objects.petri_net.utils.petri_utils import construct_trace_net
    from pm4py.objects.petri_net.utils.synchronous_product import construct
    from pm4py.objects.petri_net.utils.align_utils import SKIP

    trace_net, trace_im, trace_fm = construct_trace_net(trace)
    sync_net, sync_im, sync_fm = construct(
        trace_net,
        trace_im,
        trace_fm,
        petri_net,
        initial_marking,
        final_marking,
        SKIP,
    )
    return sync_net, sync_im, sync_fm


def compute_emd(
    language1: Dict[List[str], float], language2: Dict[List[str], float]
) -> float:
    """
    Computes the Earth Mover Distance (EMD) between two stochastic languages. For example, one language may be extracted from a log, and the other from a process model.

    :param language1: The first stochastic language.
    :param language2: The second stochastic language.
    :return: The computed Earth Mover Distance.
    :rtype: float

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes('tests/input_data/running-example.xes')
        language_log = pm4py.get_stochastic_language(log)
        print(language_log)
        net, im, fm = pm4py.read_pnml('tests/input_data/running-example.pnml')
        language_model = pm4py.get_stochastic_language(net, im, fm)
        print(language_model)
        emd_distance = pm4py.compute_emd(language_log, language_model)
        print(emd_distance)
    """
    from pm4py.algo.evaluation.earth_mover_distance import (
        algorithm as earth_mover_distance,
    )

    return earth_mover_distance.apply(language1, language2)


def solve_marking_equation(
    petri_net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    cost_function: Dict[PetriNet.Transition, float] = None,
) -> float:
    """
    Solves the marking equation of a Petri net using an Integer Linear Programming (ILP) approach. An optional transition-based cost function can be provided to minimize the solution.

    :param petri_net: The Petri net.
    :param initial_marking: The initial marking of the Petri net.
    :param final_marking: The final marking of the Petri net.
    :param cost_function: (Optional) A dictionary mapping transitions to their associated costs. If not provided, a default cost of 1 is assigned to each transition.
    :return: The heuristic value obtained by solving the marking equation.
    :rtype: float

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        heuristic = pm4py.solve_marking_equation(net, im, fm)
    """
    from pm4py.algo.analysis.marking_equation import (
        algorithm as marking_equation,
    )

    if cost_function is None:
        cost_function = {t: 1 for t in petri_net.transitions}

    me = marking_equation.build(
        petri_net,
        initial_marking,
        final_marking,
        parameters={"costs": cost_function},
    )
    return marking_equation.get_h_value(me)


@deprecation.deprecated(
    deprecated_in="2.3.0",
    removed_in="3.0.0",
    details="this method will be removed in a future release.",
)
def solve_extended_marking_equation(
    trace: Trace,
    sync_net: PetriNet,
    sync_im: Marking,
    sync_fm: Marking,
    split_points: Optional[List[int]] = None,
) -> float:
    """
    Computes a heuristic value (an underestimation of the cost of an alignment) between a trace
    and a synchronous product net using the extended marking equation with the standard cost function.
    For example, synchronization moves have a cost of 0, invisible moves have a cost of 1,
    and other moves on the model or log have a cost of 10,000. This method provides optimal provisioning of the split points.

    :param trace: The trace to evaluate.
    :param sync_net: The synchronous product net.
    :param sync_im: The initial marking of the synchronous net.
    :param sync_fm: The final marking of the synchronous net.
    :param split_points: (Optional) The indices of the events in the trace to be used as split points. If not specified, the split points are identified automatically.
    :return: The heuristic value representing the cost underestimation.
    :rtype: float

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        log = pm4py.read_xes('log.xes')
        ext_mark_eq_heu = pm4py.solve_extended_marking_equation(log[0], net, im, fm)
    """
    from pm4py.algo.analysis.extended_marking_equation import (
        algorithm as extended_marking_equation,
    )

    parameters = {}
    if split_points is not None:
        parameters[
            extended_marking_equation.Variants.CLASSIC.value.Parameters.SPLIT_IDX
        ] = split_points
    me = extended_marking_equation.build(
        trace, sync_net, sync_im, sync_fm, parameters=parameters
    )
    return extended_marking_equation.get_h_value(me)


def check_is_sound(petri_net: PetriNet,
                   initial_marking: Marking,
                   final_marking: Marking) -> bool:
    """
    Checks if a given Petri net is a sound Workflow net (WF-net).
    Returns a boolean value.

    A Petri net is a WF-net if and only if:
        - It has a unique source place.
        - It has a unique end place.
        - Every element in the WF-net is on a path from the source to the sink place.

    A WF-net is sound if and only if:
        - It contains no live-locks.
        - It contains no deadlocks.
        - It is always possible to reach the final marking from any reachable marking.

    :param petri_net: The Petri net to check.
    :param initial_marking: The initial marking of the Petri net.
    :param final_marking: The final marking of the Petri net.
    :returns: boolean (True if the Petri net is sound)
    """
    try:
        from pm4py.convert import convert_to_powl
        powl_model = convert_to_powl(petri_net, initial_marking, final_marking)
        return True
    except:
        pass

    from pm4py.algo.analysis.woflan import algorithm as woflan
    soundness = woflan.apply(
        petri_net,
        initial_marking,
        final_marking,
        parameters={
            "return_asap_when_not_sound": True,
            "return_diagnostics": True,
            "print_diagnostics": False,
        },
    )

    return soundness[0]


@deprecation.deprecated(
    deprecated_in="2.3.0",
    removed_in="3.0.0",
    details="this method will be removed in a future release.",
)
def check_soundness(
    petri_net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    print_diagnostics: bool = False,
) -> Tuple[bool, Dict[str, Any]]:
    """
    Checks if a given Petri net is a sound Workflow net (WF-net).

    A Petri net is a WF-net if and only if:
        - It has a unique source place.
        - It has a unique end place.
        - Every element in the WF-net is on a path from the source to the sink place.

    A WF-net is sound if and only if:
        - It contains no live-locks.
        - It contains no deadlocks.
        - It is always possible to reach the final marking from any reachable marking.

    For a formal definition of a sound WF-net, refer to: http://www.padsweb.rwth-aachen.de/wvdaalst/publications/p628.pdf

    The returned tuple consists of:
        - A boolean indicating whether the Petri net is a sound WF-net.
        - A dictionary containing diagnostics collected while running WOFLAN, associating diagnostic names with their corresponding details.

    :param petri_net: The Petri net to check.
    :param initial_marking: The initial marking of the Petri net.
    :param final_marking: The final marking of the Petri net.
    :param print_diagnostics: If True, additional diagnostics will be printed during the execution of WOFLAN.
    :return: A tuple containing a boolean indicating soundness and a dictionary of diagnostics.
    :rtype: Tuple[bool, Dict[str, Any]]

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        is_sound = pm4py.check_soundness(net, im, fm)
    """
    from pm4py.algo.analysis.woflan import algorithm as woflan

    return woflan.apply(
        petri_net,
        initial_marking,
        final_marking,
        parameters={
            "return_asap_when_not_sound": True,
            "return_diagnostics": True,
            "print_diagnostics": print_diagnostics,
        },
    )


def cluster_log(
    log: Union[EventLog, EventStream, pd.DataFrame],
    sklearn_clusterer=None,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
) -> Generator[EventLog, None, None]:
    """
    Applies clustering to the provided event log by extracting profiles for the log's traces and clustering them using a Scikit-Learn clusterer (default is K-Means with two clusters).

    :param log: The event log to cluster.
    :param sklearn_clusterer: (Optional) The Scikit-Learn clusterer to use. Default is KMeans with `n_clusters=2`, `random_state=0`, and `n_init="auto"`.
    :param activity_key: The key used to identify activities in the log.
    :param timestamp_key: The key used to identify timestamps in the log.
    :param case_id_key: The key used to identify case IDs in the log.
    :return: A generator that yields clustered event logs as pandas DataFrames.
    :rtype: Generator[pd.DataFrame, None, None]

    .. code-block:: python3

        import pm4py

        for clust_log in pm4py.cluster_log(df):
            print(clust_log)
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )
    if sklearn_clusterer is not None:
        properties["sklearn_clusterer"] = sklearn_clusterer

    from pm4py.algo.clustering.profiles import algorithm as clusterer

    return clusterer.apply(log, parameters=properties)


def insert_artificial_start_end(
    log: Union[EventLog, pd.DataFrame],
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    artificial_start=constants.DEFAULT_ARTIFICIAL_START_ACTIVITY,
    artificial_end=constants.DEFAULT_ARTIFICIAL_END_ACTIVITY,
) -> Union[EventLog, pd.DataFrame]:
    """
    Inserts artificial start and end activities into an event log or a Pandas DataFrame.

    :param log: The event log or Pandas DataFrame to modify.
    :param activity_key: The attribute key used for activities.
    :param timestamp_key: The attribute key used for timestamps.
    :param case_id_key: The attribute key used to identify cases.
    :param artificial_start: The symbol to use for the artificial start activity.
    :param artificial_end: The symbol to use for the artificial end activity.
    :return: The event log or Pandas DataFrame with artificial start and end activities inserted.
    :rtype: Union[EventLog, pd.DataFrame]

    .. code-block:: python3

        import pm4py

        dataframe = pm4py.insert_artificial_start_end(
            dataframe,
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )
    properties[constants.PARAM_ARTIFICIAL_START_ACTIVITY] = artificial_start
    properties[constants.PARAM_ARTIFICIAL_END_ACTIVITY] = artificial_end

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(
            log,
            activity_key=activity_key,
            case_id_key=case_id_key,
            timestamp_key=timestamp_key,
        )
        from pm4py.objects.log.util import dataframe_utils

        return dataframe_utils.insert_artificial_start_end(
            log, parameters=properties
        )
    else:
        from pm4py.objects.log.util import artificial

        return artificial.insert_artificial_start_end(
            log, parameters=properties
        )


def insert_case_service_waiting_time(
    log: Union[EventLog, pd.DataFrame],
    service_time_column: str = "@@service_time",
    sojourn_time_column: str = "@@sojourn_time",
    waiting_time_column: str = "@@waiting_time",
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    start_timestamp_key: str = "time:timestamp",
) -> pd.DataFrame:
    """
    Inserts service time, waiting time, and sojourn time information for each case into a Pandas DataFrame.

    :param log: The event log or Pandas DataFrame to modify.
    :param service_time_column: The name of the column to store service times.
    :param sojourn_time_column: The name of the column to store sojourn times.
    :param waiting_time_column: The name of the column to store waiting times.
    :param activity_key: The attribute key used for activities.
    :param timestamp_key: The attribute key used for timestamps.
    :param case_id_key: The attribute key used to identify cases.
    :param start_timestamp_key: The attribute key used for the start timestamp of cases.
    :return: A Pandas DataFrame with the inserted service, waiting, and sojourn time columns.
    :rtype: pd.DataFrame

    .. code-block:: python3

        import pm4py

        dataframe = pm4py.insert_case_service_waiting_time(
            dataframe,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name',
            start_timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )

    from pm4py.objects.conversion.log import converter as log_converter

    log_df = log_converter.apply(
        log,
        variant=log_converter.Variants.TO_DATA_FRAME,
        parameters=properties,
    )

    return pandas_utils.insert_case_service_waiting_time(
        log_df,
        case_id_column=case_id_key,
        timestamp_column=timestamp_key,
        start_timestamp_column=start_timestamp_key,
        service_time_column=service_time_column,
        waiting_time_column=waiting_time_column,
        sojourn_time_column=sojourn_time_column,
    )


def insert_case_arrival_finish_rate(
    log: Union[EventLog, pd.DataFrame],
    arrival_rate_column: str = "@@arrival_rate",
    finish_rate_column: str = "@@finish_rate",
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    start_timestamp_key: str = "time:timestamp",
) -> pd.DataFrame:
    """
    Inserts arrival and finish rate information for each case into a Pandas DataFrame.

    The arrival rate is computed as the time difference between the start of the current case and the start of the previous case to start.
    The finish rate is computed as the time difference between the end of the current case and the end of the next case to finish.

    :param log: The event log or Pandas DataFrame to modify.
    :param arrival_rate_column: The name of the column to store arrival rates.
    :param finish_rate_column: The name of the column to store finish rates.
    :param activity_key: The attribute key used for activities.
    :param timestamp_key: The attribute key used for timestamps.
    :param case_id_key: The attribute key used to identify cases.
    :param start_timestamp_key: The attribute key used for the start timestamp of cases.
    :return: A Pandas DataFrame with the inserted arrival and finish rate columns.
    :rtype: pd.DataFrame

    .. code-block:: python3

        import pm4py

        dataframe = pm4py.insert_case_arrival_finish_rate(
            dataframe,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name',
            start_timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(
        log,
        activity_key=activity_key,
        case_id_key=case_id_key,
        timestamp_key=timestamp_key,
    )

    from pm4py.objects.conversion.log import converter as log_converter

    log_df = log_converter.apply(
        log,
        variant=log_converter.Variants.TO_DATA_FRAME,
        parameters=properties,
    )

    return pandas_utils.insert_case_arrival_finish_rate(
        log_df,
        case_id_column=case_id_key,
        timestamp_column=timestamp_key,
        start_timestamp_column=start_timestamp_key,
        arrival_rate_column=arrival_rate_column,
        finish_rate_column=finish_rate_column,
    )


def check_is_workflow_net(net: PetriNet) -> bool:
    """
    Checks if the input Petri net satisfies the WF-net (Workflow net) conditions:
    1. It has a unique source place.
    2. It has a unique sink place.
    3. Every node is on a path from the source to the sink.

    :param net: The Petri net to check.
    :return: True if the Petri net is a WF-net, False otherwise.
    :rtype: bool

    .. code-block:: python3

        import pm4py

        net = pm4py.read_pnml('model.pnml')
        is_wfnet = pm4py.check_is_workflow_net(net)
    """
    from pm4py.algo.analysis.workflow_net import algorithm

    return algorithm.apply(net)


def maximal_decomposition(
    net: PetriNet, im: Marking, fm: Marking
) -> List[Tuple[PetriNet, Marking, Marking]]:
    """
    Calculates the maximal decomposition of an accepting Petri net into its maximal components.

    :param net: The Petri net to decompose.
    :param im: The initial marking of the Petri net.
    :param fm: The final marking of the Petri net.
    :return: A list of tuples, each containing a subnet Petri net, its initial marking, and its final marking.
    :rtype: List[Tuple[PetriNet, Marking, Marking]]

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        list_nets = pm4py.maximal_decomposition(net, im, fm)
        for subnet, subim, subfm in list_nets:
            pm4py.view_petri_net(subnet, subim, subfm, format='svg')
    """
    from pm4py.objects.petri_net.utils.decomposition import decompose

    return decompose(net, im, fm)


def simplicity_petri_net(
    net: PetriNet,
    im: Marking,
    fm: Marking,
    variant: Optional[str] = "arc_degree",
) -> float:
    """
    Computes the simplicity metric for a given Petri net model.

    Three available approaches are supported:
    - **Arc Degree Simplicity**: Described in the paper "ProDiGen: Mining complete, precise and minimal structure process models with a genetic algorithm." by Vázquez-Barreiros, Borja, Manuel Mucientes, and Manuel Lama. Information Sciences, 294 (2015): 315-333.
    - **Extended Cardoso Metric**: Described in the paper "Complexity Metrics for Workflow Nets" by Lassen, Kristian Bisgaard, and Wil MP van der Aalst.
    - **Extended Cyclomatic Metric**: Also described in the paper "Complexity Metrics for Workflow Nets" by Lassen, Kristian Bisgaard, and Wil MP van der Aalst.

    :param net: The Petri net for which to compute simplicity.
    :param im: The initial marking of the Petri net.
    :param fm: The final marking of the Petri net.
    :param variant: The simplicity metric variant to use ('arc_degree', 'extended_cardoso', 'extended_cyclomatic').
    :return: The computed simplicity value.
    :rtype: float

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.discover_petri_net_inductive(
            dataframe,
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
        simplicity = pm4py.simplicity_petri_net(net, im, fm, variant='arc_degree')
    """
    if variant == "arc_degree":
        from pm4py.algo.evaluation.simplicity.variants import arc_degree

        return arc_degree.apply(net)
    elif variant == "extended_cardoso":
        from pm4py.algo.evaluation.simplicity.variants import extended_cardoso

        return extended_cardoso.apply(net)
    elif variant == "extended_cyclomatic":
        from pm4py.algo.evaluation.simplicity.variants import (
            extended_cyclomatic,
        )

        return extended_cyclomatic.apply(net, im)


def generate_marking(
    net: PetriNet,
    place_or_dct_places: Union[
        str, PetriNet.Place, Dict[str, int], Dict[PetriNet.Place, int]
    ],
) -> Marking:
    """
    Generates a marking for a given Petri net based on specified places and token counts.

    :param net: The Petri net for which to generate the marking.
    :param place_or_dct_places: Specifies the places and their token counts for the marking. It can be:
        - A single `PetriNet.Place` object, which will have one token.
        - A string representing the name of a place, which will have one token.
        - A dictionary mapping `PetriNet.Place` objects to their respective number of tokens.
        - A dictionary mapping place names (strings) to their respective number of tokens.
    :return: The generated Marking object.
    :rtype: Marking

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        marking = pm4py.generate_marking(net, {'source': 2})
    """
    dct_places = {x.name: x for x in net.places}
    if isinstance(place_or_dct_places, PetriNet.Place):
        # A single Place object is specified for the marking
        return Marking({place_or_dct_places: 1})
    elif isinstance(place_or_dct_places, str):
        # The name of a place is specified for the marking
        return Marking({dct_places[place_or_dct_places]: 1})
    elif isinstance(place_or_dct_places, dict):
        dct_keys = list(place_or_dct_places)
        if dct_keys:
            if isinstance(dct_keys[0], PetriNet.Place):
                # A dictionary mapping Place objects to token counts is
                # specified
                return Marking(place_or_dct_places)
            elif isinstance(dct_keys[0], str):
                # A dictionary mapping place names to token counts is specified
                return Marking(
                    {dct_places[x]: y for x, y in place_or_dct_places.items()}
                )


def reduce_petri_net_invisibles(net: PetriNet) -> PetriNet:
    """
    Reduces the number of invisible transitions in the provided Petri net.

    :param net: The Petri net to be reduced.
    :return: The reduced Petri net with fewer invisible transitions.
    :rtype: PetriNet

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        net = pm4py.reduce_petri_net_invisibles(net)
    """
    from pm4py.objects.petri_net.utils import reduction

    return reduction.apply_simple_reduction(net)


def reduce_petri_net_implicit_places(
    net: PetriNet, im: Marking, fm: Marking
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Reduces the number of implicit places in the provided Petri net.

    :param net: The Petri net to be reduced.
    :param im: The initial marking of the Petri net.
    :param fm: The final marking of the Petri net.
    :return: A tuple containing the reduced Petri net, its initial marking, and its final marking.
    :rtype: Tuple[PetriNet, Marking, Marking]

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('model.pnml')
        net, im, fm = pm4py.reduce_petri_net_implicit_places(net, im, fm)
    """
    from pm4py.objects.petri_net.utils import murata

    return murata.apply_reduction(net, im, fm)


def get_enabled_transitions(
    net: PetriNet, marking: Marking
) -> Set[PetriNet.Transition]:
    """
    Retrieves the set of transitions that are enabled in a given marking of a Petri net.

    :param net: The Petri net.
    :param marking: The current marking of the Petri net.
    :return: A set of transitions that are enabled in the provided marking.
    :rtype: Set[PetriNet.Transition]

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.read_pnml('tests/input_data/running-example.pnml')
        # Gets the transitions enabled in the initial marking
        enabled_transitions = pm4py.get_enabled_transitions(net, im)
    """
    from pm4py.objects.petri_net import semantics

    return semantics.enabled_transitions(net, marking)


def get_activity_labels(*args) -> List[str]:
    """
    Gets the activity labels from the specified event log / process model.

    Returns
    ---------------
    activities
        Activity labels
    """
    import pm4py

    if isinstance(args[0], EventLog):
        labels = set(y["concept:name"] for x in args[0] for y in x)
    elif isinstance(args[0], pd.DataFrame):
        labels = set(args[0]["concept:name"].unique())
    else:
        net, im, fm = pm4py.convert_to_petri_net(*args)
        labels = {x.label for x in net.transitions if x.label is not None}
    return sorted(list(labels))


def replace_activity_labels(string_dictio, *args):
    """
    Replace the activity labels in the specified process model.
    The first argument is the dictionary, i.e., {"pay": "pay compensation", "reject": "reject request"}
    The rest is the specification of the process model
    """
    from pm4py.objects.powl.obj import POWL
    from pm4py.objects.bpmn.obj import BPMN

    if isinstance(args[0], POWL):
        from pm4py.objects.powl.utils import label_replacing
        return label_replacing.apply(args[0], string_dictio)
    elif isinstance(args[0], ProcessTree):
        from pm4py.objects.process_tree.utils import label_replacing
        return label_replacing.apply(args[0], string_dictio)
    elif isinstance(args[0], PetriNet):
        from pm4py.objects.petri_net.utils import label_replacing
        return label_replacing.apply(args[0], args[1], args[2], string_dictio)
    elif isinstance(args[0], BPMN):
        from pm4py.objects.bpmn.util import label_replacing
        return label_replacing.apply(args[0], string_dictio)
    else:
        raise Exception("unsupported.")


def __extract_models(*args) -> List[Any]:
    if len(args) < 2:
        raise Exception("Insufficient arguments provided.")

    counter = 0
    lst_models = []

    import pm4py
    for i in range(2):
        if type(args[counter]) is PetriNet:
            net, im, fm = args[counter:counter + 3]
            lst_models.append([net, im, fm])
            counter += 3
        elif isinstance(args[counter], dict):
            dfg, sa, ea = args[counter:counter + 3]
            net, im, fm = pm4py.convert_to_petri_net(dfg, sa, ea)
            lst_models.append([net, im, fm])
            counter += 3
        else:
            obj = args[counter]
            lst_models.append([obj])
            counter += 1

    return lst_models


def behavioral_similarity(*args) -> float:
    """
    Computes the behavioral similarity (footprints-based) between two process models.

    Examples:
    * pm4py.behavioral_similarity(petri_net, im, fm, process_tree)
    * pm4py.behavioral_similarity(bpmn1, bpmn2)
    * pm4py.behavioral_similarity(process_tree, powl)

    Returns
    --------------
    similarity
        Footprints-based behavioral similarity
    """
    lst_models = __extract_models(*args)

    import pm4py
    footprints = []
    for i in range(len(lst_models)):
        x = lst_models[i]
        if not (isinstance(x[0], PetriNet) or isinstance(x[0], ProcessTree)):
            x = [pm4py.convert_to_powl(*x)]

        footprints.append(pm4py.discover_footprints(*x))

    footprints1, footprints2 = footprints

    sequence_union = footprints1["sequence"].union(footprints2["sequence"])
    sequence_intersection = footprints1["sequence"].intersection(footprints2["sequence"])

    parallel_union = footprints1["parallel"].union(footprints2["parallel"])
    parallel_intersection = footprints1["parallel"].intersection(footprints2["parallel"])

    denominator = len(sequence_union) + len(parallel_union)

    if denominator == 0:
        return 0
    else:
        return (len(sequence_intersection) + len(parallel_intersection)) / denominator


def structural_similarity(*args) -> float:
    """
    Computes the structural similarity between two semi-block-structured process models,
    following an approach similar to:

    Yan, Z., Dijkman, R., & Grefen, P. (2012). Fast business process similarity search.
    Distributed and Parallel Databases, 30(2), 105–144.
    (https://doi.org/10.1007/s10619-012-7089-z)

    Examples:
    * pm4py.structural_similarity(petri_net, im, fm, process_tree)
    * pm4py.structural_similarity(bpmn1, bpmn2)
    * pm4py.structural_similarity(process_tree, powl)

    Returns
    --------------
    similarity
        Structural similarity
    """
    lst_models = __extract_models(*args)

    import pm4py
    i = 0
    while i < len(lst_models):
        lst_models[i] = pm4py.convert_to_process_tree(pm4py.convert_to_powl(*lst_models[i]))
        i = i + 1

    from pm4py.objects.process_tree.utils import struct_similarity
    return struct_similarity.structural_similarity(lst_models[0], lst_models[1])


def embeddings_similarity(*args) -> float:
    """
    Computes the embeddings similarity between two process models,
    following the approach described in:

    Colonna, Juan G., et al. "Process mining embeddings: Learning vector representations for Petri nets."
    Intelligent Systems with Applications 23 (2024): 200423.

    Examples:
    * pm4py.embeddings_similarity(petri_net, im, fm, process_tree)
    * pm4py.embeddings_similarity(bpmn1, bpmn2)
    * pm4py.embeddings_similarity(process_tree, powl)

    Returns
    --------------
    similarity
        Structural similarity
    """
    lst_models = __extract_models(*args)

    import pm4py
    i = 0
    while i < len(lst_models):
        lst_models[i] = pm4py.convert_to_petri_net(*lst_models[i])
        i = i + 1

    from pm4py.objects.petri_net.utils import embeddings_similarity
    return embeddings_similarity.apply(lst_models[0][0], lst_models[1][0])


def label_sets_similarity(*args, threshold=0.75) -> float:
    """
    Computes the label sets similarity between two process models.

    Examples:
    * pm4py.labels_similarity(petri_net, im, fm, process_tree)
    * pm4py.labels_similarity(bpmn1, bpmn2)
    * pm4py.labels_similarity(process_tree, powl)

    Returns
    --------------
    similarity
        Label sets similarity
    """
    lst_models = __extract_models(*args)
    labels = []
    i = 0
    while i < len(lst_models):
        labels.append(get_activity_labels(*lst_models[i]))
        i = i + 1

    return ls_util.label_sets_similarity(labels[0], labels[1], threshold=threshold)


def map_labels_from_second_model(*args, threshold=0.75):
    """
    Maps the labels from the second process model into the first.

    Example usages:
    * pm4py.map_labels_from_second_model(net, im, fm, process_tree)
    * pm4py.map_labels_from_second_model(process_tree, net, im, fm)
    * pm4py.map_labels_from_second_model(powl1, powl2)
    """
    lst_models = __extract_models(*args)
    labels = []
    i = 0
    while i < len(lst_models):
        labels.append(get_activity_labels(*lst_models[i]))
        i = i + 1

    label_mapping = ls_util.map_labels(labels[0], labels[1], threshold=threshold)
    return replace_activity_labels(label_mapping, *lst_models[0])



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import math
from datetime import timedelta, datetime, time
from typing import List, Tuple

from pm4py.util import constants
from pm4py.util.dt_parsing.variants import strpfromiso


class BusinessHours:
    def __init__(self, datetime1, datetime2, **kwargs):
        # Remove timezone info for simplicity (assumes same timezone)
        self.datetime1 = datetime1.replace(tzinfo=None)
        self.datetime2 = datetime2.replace(tzinfo=None)
        # Use provided business hour slots or default
        self.business_hour_slots = (
            kwargs["business_hour_slots"]
            if "business_hour_slots" in kwargs
            else constants.DEFAULT_BUSINESS_HOUR_SLOTS
        )
        # Unify slots to avoid overlaps
        self.business_hour_slots_unified = []
        for begin, end in sorted(self.business_hour_slots):
            if self.business_hour_slots_unified and self.business_hour_slots_unified[-1][1] >= begin - 1:
                self.business_hour_slots_unified[-1][1] = max(self.business_hour_slots_unified[-1][1], end)
            else:
                self.business_hour_slots_unified.append([begin, end])
        # Work calendar (unused in this implementation)
        self.work_calendar = (
            kwargs["work_calendar"]
            if "work_calendar" in kwargs
            else constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR
        )

    def business_seconds_from_week_start(self, dt):
        """Calculate business seconds from the start of the week to the given datetime."""
        week_start = dt.date() - timedelta(days=dt.weekday())  # Monday 00:00
        seconds_since_week_start = (dt - datetime.combine(week_start, time.min)).total_seconds()
        sum_overlap = 0
        for start, end in self.business_hour_slots_unified:
            # Overlap is max(0, min(end, seconds) - start)
            sum_overlap += max(0, min(seconds_since_week_start, end) - start)
        return sum_overlap

    def get_seconds(self):
        """Calculate total business seconds between datetime1 and datetime2."""
        if self.datetime2 <= self.datetime1:
            return 0.0

        # Total business seconds in a full week
        total_business_seconds_per_week = sum(end - start for start, end in self.business_hour_slots_unified)

        # Week starts (Monday 00:00)
        week_start1 = self.datetime1.date() - timedelta(days=self.datetime1.weekday())
        week_start2 = self.datetime2.date() - timedelta(days=self.datetime2.weekday())

        # Number of weeks between week starts
        number_of_weeks = (week_start2 - week_start1).days // 7

        # Business seconds from week start to each datetime
        s1 = self.business_seconds_from_week_start(self.datetime1)
        s2 = self.business_seconds_from_week_start(self.datetime2)

        # Total = full weeks + partial week at end - partial week at start
        total = total_business_seconds_per_week * number_of_weeks + s2 - s1
        return total


def soj_time_business_hours_diff(
    st: datetime,
    et: datetime,
    business_hour_slots: List[Tuple[int]],
    work_calendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
) -> float:
    """
    Calculates the difference between the provided timestamps based on business hours.

    Parameters
    ----------
    st : datetime
        Start timestamp
    et : datetime
        End timestamp
    business_hour_slots : List[Tuple[int]]
        Work schedule as list of tuples (start, end) in seconds since week start
    work_calendar
        Work calendar (unused in this implementation)

    Returns
    -------
    float
        Difference in business hours (seconds)
    """
    bh = BusinessHours(
        st,
        et,
        business_hour_slots=business_hour_slots,
        work_calendar=work_calendar,
    )
    return bh.get_seconds()


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import collections
import threading
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from pm4py.util import exec_utils


class StreamState(Enum):
    INACTIVE = 1
    ACTIVE = 2
    FINISHED = 3


class Parameters(Enum):
    THREAD_POOL_SIZE = "thread_pool_size"


class LiveEventStream:

    def __init__(self, parameters=None):
        self._dq = collections.deque()
        self._state = StreamState.INACTIVE
        self._lock = threading.Lock()
        self._cond = threading.Condition(self._lock)
        self._observers = set()
        self._mail_man = None
        self._tp = ThreadPoolExecutor(
            exec_utils.get_param_value(
                Parameters.THREAD_POOL_SIZE, parameters, 6
            )
        )

    def append(self, event):
        self._cond.acquire()
        if self._state != StreamState.FINISHED:
            self._dq.append(event)
            self._cond.notify()
        self._cond.release()

    def _deliver(self):
        while self._state != StreamState.INACTIVE:
            self._cond.acquire()
            while len(self._dq) == 0:
                self._cond.notify()
                if self._state != StreamState.FINISHED:
                    self._cond.wait()
                else:
                    self._cond.release()
                    return
            event = self._dq.popleft()
            for algo in self._observers:
                self._tp.submit(algo.receive, event)
            self._cond.release()

    def start(self):
        self._cond.acquire()
        self._state = StreamState.ACTIVE
        self._mail_man = threading.Thread(target=self._deliver)
        self._mail_man.start()
        self._cond.release()

    def stop(self):
        self._cond.acquire()
        while len(self._dq) > 0:
            self._cond.wait()
        self._tp.shutdown()
        if self._state == StreamState.ACTIVE:
            self._state = StreamState.FINISHED
            self._cond.notify()
        self._cond.release()

    def register(self, algo):
        self._cond.acquire()
        self._observers.add(algo)
        self._cond.release()

    async def deregister(self, algo):
        self._cond.acquire()
        self._observers.remove(algo)
        self._cond.release()

    def _get_state(self):
        return self._state

    state = property(_get_state)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from numpy.random import choice


def pick_transition(et, smap):
    """
    Pick a transition in a set of transitions based on the weights
    specified by the stochastic map

    Parameters
    --------------
    et
        Enabled transitions
    smap
        Stochastic map

    Returns
    --------------
    trans
        Transition chosen according to the weights
    """
    wmap = {ct: smap[ct].get_weight() if ct in smap else 1.0 for ct in et}
    wmap_sv = sum(wmap.values())
    list_of_candidates = []
    probability_distribution = []
    for ct in wmap:
        list_of_candidates.append(ct)
        if wmap_sv == 0:
            probability_distribution.append(1.0 / float(len(wmap)))
        else:
            probability_distribution.append(wmap[ct] / wmap_sv)
    ct = list(choice(et, 1, p=probability_distribution))[0]
    return ct



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
"""

import datetime
from typing import Optional, Tuple, Any, Collection, Union, List

import pandas as pd

from pm4py.objects.log.obj import EventLog, EventStream, Trace
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.powl.obj import POWL
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import constants, xes_constants, pandas_utils
import warnings
from pm4py.util.pandas_utils import (
    check_is_pandas_dataframe,
    check_pandas_dataframe_columns,
)
from pm4py.util.dt_parsing.variants import strpfromiso
from pm4py.util import deprecation


INDEX_COLUMN = "@@index"
CASE_INDEX_COLUMN = "@@case_index"


class Shared:
    RUSTXES_WARNING_SHOWN = False
    FILTERING_LEVEL_WARNING_SHOWN = False


def is_polars_lazyframe(df: Any) -> bool:
    """Return True if the provided dataframe is a Polars LazyFrame."""
    df_type = str(type(df)).lower()
    return "polars" in df_type and "lazyframe" in df_type


def format_dataframe(
    df,
    case_id: str = constants.CASE_CONCEPT_NAME,
    activity_key: str = xes_constants.DEFAULT_NAME_KEY,
    timestamp_key: str = xes_constants.DEFAULT_TIMESTAMP_KEY,
    start_timestamp_key: str = xes_constants.DEFAULT_START_TIMESTAMP_KEY,
    timest_format: Optional[str] = None,
) -> Any:
    """
    Formats the dataframe appropriately for process mining purposes.

    :param df: Dataframe.
    :param case_id: Case identifier column.
    :param activity_key: Activity column.
    :param timestamp_key: Timestamp column.
    :param start_timestamp_key: Start timestamp column.
    :param timest_format: Timestamp format provided to Pandas.
    :return: A formatted dataframe (Pandas DataFrame or Polars LazyFrame).
    :rtype: Any

    .. code-block:: python3

        import pandas as pd
        import pm4py

        dataframe = pd.read_csv('event_log.csv')
        dataframe = pm4py.format_dataframe(
            dataframe,
            case_id='case:concept:name',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            start_timestamp_key='start_timestamp',
            timest_format='%Y-%m-%d %H:%M:%S'
        )
    """
    if timest_format is None:
        timest_format = constants.DEFAULT_TIMESTAMP_PARSE_FORMAT

    if is_polars_lazyframe(df):
        import polars as pl  # type: ignore[import-untyped]

        lf = df
        schema = lf.collect_schema()
        column_names = schema.names()
        required_columns = (case_id, activity_key, timestamp_key)
        for col in required_columns:
            if col not in column_names:
                raise Exception(f"{col} column is not in the dataframe!")

        def refresh_schema():
            nonlocal lf, schema, column_names
            schema = lf.collect_schema()
            column_names = schema.names()

        def drop_if_present(col_name):
            nonlocal lf
            if col_name in column_names:
                lf = lf.drop(col_name)
                refresh_schema()

        if (
            case_id != constants.CASE_CONCEPT_NAME
            or constants.CASE_CONCEPT_NAME not in column_names
        ):
            drop_if_present(constants.CASE_CONCEPT_NAME)
            lf = lf.with_columns(
                pl.col(case_id).alias(constants.CASE_CONCEPT_NAME)
            )
            refresh_schema()
        if (
            activity_key != xes_constants.DEFAULT_NAME_KEY
            or xes_constants.DEFAULT_NAME_KEY not in column_names
        ):
            drop_if_present(xes_constants.DEFAULT_NAME_KEY)
            lf = lf.with_columns(
                pl.col(activity_key).alias(xes_constants.DEFAULT_NAME_KEY)
            )
            refresh_schema()
        if (
            timestamp_key != xes_constants.DEFAULT_TIMESTAMP_KEY
            or xes_constants.DEFAULT_TIMESTAMP_KEY not in column_names
        ):
            drop_if_present(xes_constants.DEFAULT_TIMESTAMP_KEY)
            lf = lf.with_columns(
                pl.col(timestamp_key).alias(xes_constants.DEFAULT_TIMESTAMP_KEY)
            )
            refresh_schema()

        if start_timestamp_key in column_names:
            lf = lf.with_columns(
                pl.col(start_timestamp_key).alias(
                    xes_constants.DEFAULT_START_TIMESTAMP_KEY
                )
            )
            refresh_schema()

        timestamp_targets = [
            col
            for col in (
                xes_constants.DEFAULT_TIMESTAMP_KEY,
                xes_constants.DEFAULT_START_TIMESTAMP_KEY,
            )
            if col in column_names
        ]

        schema = lf.collect_schema()
        for col in timestamp_targets:
            dtype = schema.get(col)
            expr = pl.col(col)
            if dtype != pl.Datetime:
                expr = expr.cast(pl.Utf8).str.strptime(
                    pl.Datetime,
                    format=timest_format,
                    strict=False,
                    exact=False,
                )
            expr = expr.dt.replace_time_zone("UTC")
            lf = lf.with_columns(expr.alias(col))
        refresh_schema()

        lf = lf.filter(
            pl.col(constants.CASE_CONCEPT_NAME).is_not_null()
            & pl.col(xes_constants.DEFAULT_NAME_KEY).is_not_null()
            & pl.col(xes_constants.DEFAULT_TIMESTAMP_KEY).is_not_null()
        )

        lf = lf.with_columns(
            pl.col(constants.CASE_CONCEPT_NAME).cast(pl.Utf8),
            pl.col(xes_constants.DEFAULT_NAME_KEY).cast(pl.Utf8),
        )

        lf = pandas_utils.insert_index(
            lf, INDEX_COLUMN, copy_dataframe=False
        )
        lf = lf.sort(
            [
                constants.CASE_CONCEPT_NAME,
                xes_constants.DEFAULT_TIMESTAMP_KEY,
                INDEX_COLUMN,
            ]
        )
        lf = pandas_utils.insert_index(
            lf, INDEX_COLUMN, copy_dataframe=False
        )
        lf = pandas_utils.insert_case_index(
            lf,
            CASE_INDEX_COLUMN,
            copy_dataframe=False,
        )
        lf = lf.collect().lazy()
        return lf

    from pm4py.objects.log.util import dataframe_utils

    if case_id not in df.columns:
        raise Exception(case_id + " column (case ID) is not in the dataframe!")
    if activity_key not in df.columns:
        raise Exception(
            activity_key + " column (activity) is not in the dataframe!"
        )
    if timestamp_key not in df.columns:
        raise Exception(
            timestamp_key + " column (timestamp) is not in the dataframe!"
        )
    if case_id != constants.CASE_CONCEPT_NAME:
        if constants.CASE_CONCEPT_NAME in df.columns:
            del df[constants.CASE_CONCEPT_NAME]
        df[constants.CASE_CONCEPT_NAME] = df[case_id]
    if activity_key != xes_constants.DEFAULT_NAME_KEY:
        if xes_constants.DEFAULT_NAME_KEY in df.columns:
            del df[xes_constants.DEFAULT_NAME_KEY]
        df[xes_constants.DEFAULT_NAME_KEY] = df[activity_key]
    if timestamp_key != xes_constants.DEFAULT_TIMESTAMP_KEY:
        if xes_constants.DEFAULT_TIMESTAMP_KEY in df.columns:
            del df[xes_constants.DEFAULT_TIMESTAMP_KEY]
        df[xes_constants.DEFAULT_TIMESTAMP_KEY] = df[timestamp_key]
    # Makes sure that the timestamps column are of timestamp type
    df = dataframe_utils.convert_timestamp_columns_in_df(
        df, timest_format=timest_format
    )
    # Drop NaN(s) in the main columns (case ID, activity, timestamp) to ensure
    # functioning of the algorithms
    prev_length = len(df)
    df = df.dropna(
        subset={
            constants.CASE_CONCEPT_NAME,
            xes_constants.DEFAULT_NAME_KEY,
            xes_constants.DEFAULT_TIMESTAMP_KEY,
        },
        how="any",
    )

    if len(df) < prev_length:
        if constants.SHOW_INTERNAL_WARNINGS:
            warnings.warn(
                "Some rows of the Pandas data frame have been removed because of empty case IDs, activity labels, or timestamps to ensure the correct functioning of PM4Py's algorithms."
            )

    # Make sure the case ID column is of string type
    df[constants.CASE_CONCEPT_NAME] = df[constants.CASE_CONCEPT_NAME].astype(
        "string"
    )
    # Make sure the activity column is of string type
    df[xes_constants.DEFAULT_NAME_KEY] = df[
        xes_constants.DEFAULT_NAME_KEY
    ].astype("string")
    # Set an index column
    df = pandas_utils.insert_index(df, INDEX_COLUMN, copy_dataframe=False)
    # Sorts the dataframe
    df = df.sort_values(
        [
            constants.CASE_CONCEPT_NAME,
            xes_constants.DEFAULT_TIMESTAMP_KEY,
            INDEX_COLUMN,
        ]
    )
    # Re-set the index column
    df = pandas_utils.insert_index(df, INDEX_COLUMN, copy_dataframe=False)
    # Sets the index column in the dataframe
    df = pandas_utils.insert_case_index(
        df, CASE_INDEX_COLUMN, copy_dataframe=False
    )
    # Sets the properties
    if not hasattr(df, "attrs"):
        # Legacy (Python 3.6) support
        df.attrs = {}
    if start_timestamp_key in df.columns:
        df[xes_constants.DEFAULT_START_TIMESTAMP_KEY] = df[start_timestamp_key]
        df.attrs[constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY] = (
            xes_constants.DEFAULT_START_TIMESTAMP_KEY
        )
    df.attrs[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] = (
        xes_constants.DEFAULT_NAME_KEY
    )
    df.attrs[constants.PARAMETER_CONSTANT_TIMESTAMP_KEY] = (
        xes_constants.DEFAULT_TIMESTAMP_KEY
    )
    df.attrs[constants.PARAMETER_CONSTANT_GROUP_KEY] = (
        xes_constants.DEFAULT_GROUP_KEY
    )
    df.attrs[constants.PARAMETER_CONSTANT_TRANSITION_KEY] = (
        xes_constants.DEFAULT_TRANSITION_KEY
    )
    df.attrs[constants.PARAMETER_CONSTANT_RESOURCE_KEY] = (
        xes_constants.DEFAULT_RESOURCE_KEY
    )
    df.attrs[constants.PARAMETER_CONSTANT_CASEID_KEY] = (
        constants.CASE_CONCEPT_NAME
    )
    return df


def rebase(
    log_obj: Union[EventLog, EventStream, pd.DataFrame],
    case_id: str = constants.CASE_CONCEPT_NAME,
    activity_key: str = xes_constants.DEFAULT_NAME_KEY,
    timestamp_key: str = xes_constants.DEFAULT_TIMESTAMP_KEY,
    start_timestamp_key: str = xes_constants.DEFAULT_START_TIMESTAMP_KEY,
    timest_format: Optional[str] = None,
) -> Union[EventLog, EventStream, pd.DataFrame]:
    """
    Re-bases the log object by changing the case ID, activity, and timestamp attributes.

    :param log_obj: Log object.
    :param case_id: Case identifier.
    :param activity_key: Activity.
    :param timestamp_key: Timestamp.
    :param start_timestamp_key: Start timestamp.
    :param timest_format: Timestamp format provided to Pandas.
    :return: A re-based log object.
    :rtype: Union[EventLog, EventStream, pd.DataFrame].

    .. code-block:: python3

        import pm4py

        rebased_dataframe = pm4py.rebase(
            dataframe,
            case_id='case:concept:name',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            start_timestamp_key='start_timestamp',
            timest_format='%Y-%m-%d %H:%M:%S'
        )
    """
    import pm4py

    __event_log_deprecation_warning(log_obj)

    if is_polars_lazyframe(log_obj):
        check_pandas_dataframe_columns(log_obj)
    elif check_is_pandas_dataframe(log_obj):
        check_pandas_dataframe_columns(log_obj)

    if is_polars_lazyframe(log_obj):
        return format_dataframe(
            log_obj,
            case_id=case_id,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            start_timestamp_key=start_timestamp_key,
            timest_format=timest_format,
        )
    elif check_is_pandas_dataframe(log_obj):
        return format_dataframe(
            log_obj,
            case_id=case_id,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            start_timestamp_key=start_timestamp_key,
            timest_format=timest_format,
        )
    elif isinstance(log_obj, EventLog):
        log_obj = pm4py.convert_to_dataframe(log_obj)
        log_obj = format_dataframe(
            log_obj,
            case_id=case_id,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            start_timestamp_key=start_timestamp_key,
            timest_format=timest_format,
        )
        from pm4py.objects.conversion.log import converter

        return converter.apply(
            log_obj, variant=converter.Variants.TO_EVENT_LOG
        )
    elif isinstance(log_obj, EventStream):
        log_obj = pm4py.convert_to_dataframe(log_obj)
        log_obj = format_dataframe(
            log_obj,
            case_id=case_id,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            start_timestamp_key=start_timestamp_key,
            timest_format=timest_format,
        )
        return pm4py.convert_to_event_stream(log_obj)


def parse_process_tree(tree_string: str) -> ProcessTree:
    """
    Parses a process tree from a string.

    :param tree_string: String representing a process tree (e.g., "-> ( 'A', O ( 'B', 'C' ), 'D' )"). Operators are '->' for sequence, '+' for parallel, 'X' for XOR choice, '*' for binary loop, and 'O' for choice.
    :return: A ProcessTree object.
    :rtype: ProcessTree

    .. code-block:: python3

        import pm4py

        process_tree = pm4py.parse_process_tree("-> ( 'A', O ( 'B', 'C' ), 'D' )")
    """
    from pm4py.objects.process_tree.utils.generic import parse

    return parse(tree_string)


def parse_powl_model_string(powl_string: str) -> POWL:
    """
    Parses a POWL model from a string representation of the process model
    (with the same format as the __repr__ and __str__ methods of the POWL model).

    :param powl_string: POWL model expressed as a string (__repr__ of the POWL model).
    :return: A POWL object.
    :rtype: POWL

    .. code-block:: python3

        import pm4py

        powl_model = pm4py.parse_powl_model_string('PO=(nodes={ NODE1, NODE2, NODE3 }, order={ NODE1-->NODE2 })')
        print(powl_model)
    """
    from pm4py.objects.powl import parser

    return parser.parse_powl_model_string(powl_string)


def serialize(*args) -> Tuple[str, bytes]:
    """
    Serializes a PM4Py object into a bytes string.

    :param args: PM4Py object(s) to serialize. Supported types include:
                 - An EventLog object.
                 - A Pandas DataFrame object.
                 - A tuple consisting of (PetriNet, Marking, Marking).
                 - A ProcessTree object.
                 - A BPMN object.
                 - A DFG, including the dictionary of directly-follows relations, start activities, and end activities.
    :return: A tuple containing the serialization type as a string and the serialized bytes.
    :rtype: Tuple[str, bytes]

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.discover_petri_net_inductive(dataframe)
        serialization = pm4py.serialize(net, im, fm)
    """
    from pm4py.objects.log.obj import EventLog
    from pm4py.objects.petri_net.obj import PetriNet
    from pm4py.objects.process_tree.obj import ProcessTree
    from pm4py.objects.bpmn.obj import BPMN
    from collections import Counter

    if type(args[0]) is EventLog:
        from pm4py.objects.log.exporter.xes import exporter as xes_exporter

        return (
            constants.AvailableSerializations.EVENT_LOG.value,
            xes_exporter.serialize(*args),
        )
    elif pandas_utils.check_is_pandas_dataframe(args[0]):
        from io import BytesIO

        buffer = BytesIO()
        args[0].to_parquet(buffer)
        return (
            constants.AvailableSerializations.DATAFRAME.value,
            buffer.getvalue(),
        )
    elif len(args) == 3 and type(args[0]) is PetriNet:
        from pm4py.objects.petri_net.exporter import exporter as petri_exporter

        return (
            constants.AvailableSerializations.PETRI_NET.value,
            petri_exporter.serialize(*args),
        )
    elif type(args[0]) is ProcessTree:
        from pm4py.objects.process_tree.exporter import (
            exporter as tree_exporter,
        )

        return (
            constants.AvailableSerializations.PROCESS_TREE.value,
            tree_exporter.serialize(*args),
        )
    elif type(args[0]) is BPMN:
        from pm4py.objects.bpmn.exporter import exporter as bpmn_exporter

        return (
            constants.AvailableSerializations.BPMN.value,
            bpmn_exporter.serialize(*args),
        )
    elif len(args) == 3 and (
        isinstance(args[0], dict) or isinstance(args[0], Counter)
    ):
        from pm4py.objects.dfg.exporter import exporter as dfg_exporter

        return (
            constants.AvailableSerializations.DFG.value,
            dfg_exporter.serialize(
                args[0],
                parameters={
                    "start_activities": args[1],
                    "end_activities": args[2],
                },
            ),
        )


def deserialize(ser_obj: Tuple[str, bytes]) -> Any:
    """
    Deserializes a bytes string back into a PM4Py object.

    :param ser_obj: Serialized object as a tuple, consisting of a string indicating the type of the object and a bytes string representing the serialization.
    :return: The deserialized PM4Py object.
    :rtype: Any

    .. code-block:: python3

        import pm4py

        net, im, fm = pm4py.discover_petri_net_inductive(dataframe)
        serialization = pm4py.serialize(net, im, fm)
        net, im, fm = pm4py.deserialize(serialization)
    """
    if ser_obj[0] == constants.AvailableSerializations.EVENT_LOG.value:
        from pm4py.objects.log.importer.xes import importer as xes_importer

        return xes_importer.deserialize(ser_obj[1])
    elif ser_obj[0] == constants.AvailableSerializations.DATAFRAME.value:
        from io import BytesIO

        buffer = BytesIO()
        buffer.write(ser_obj[1])
        buffer.flush()
        return pd.read_parquet(buffer)
    elif ser_obj[0] == constants.AvailableSerializations.PETRI_NET.value:
        from pm4py.objects.petri_net.importer import importer as petri_importer

        return petri_importer.deserialize(ser_obj[1])
    elif ser_obj[0] == constants.AvailableSerializations.PROCESS_TREE.value:
        from pm4py.objects.process_tree.importer import (
            importer as tree_importer,
        )

        return tree_importer.deserialize(ser_obj[1])
    elif ser_obj[0] == constants.AvailableSerializations.BPMN.value:
        from pm4py.objects.bpmn.importer import importer as bpmn_importer

        return bpmn_importer.deserialize(ser_obj[1])
    elif ser_obj[0] == constants.AvailableSerializations.DFG.value:
        from pm4py.objects.dfg.importer import importer as dfg_importer

        return dfg_importer.deserialize(ser_obj[1])


def get_properties(
    log,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    resource_key: str = "org:resource",
    group_key: Optional[str] = None,
    start_timestamp_key: Optional[str] = None,
    **kwargs
):
    """
    Retrieves the properties from a log object.

    :param log: Log object.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param start_timestamp_key: (Optional) Attribute to be used for the start timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :param resource_key: (Optional) Attribute to be used as resource.
    :param group_key: (Optional) Attribute to be used as group identifier.
    :param kwargs: Additional keyword arguments.
    :return: A dictionary of properties.
    :rtype: Dict
    """
    __event_log_deprecation_warning(log)

    from copy import copy

    parameters = (
        copy(log.properties)
        if hasattr(log, "properties")
        else copy(log.attrs) if hasattr(log, "attrs") else {}
    )

    if activity_key is not None:
        parameters[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] = activity_key
        parameters[constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY] = activity_key

    if timestamp_key is not None:
        parameters[constants.PARAMETER_CONSTANT_TIMESTAMP_KEY] = timestamp_key

    if start_timestamp_key is not None:
        parameters[constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY] = (
            start_timestamp_key
        )

    if case_id_key is not None:
        parameters[constants.PARAMETER_CONSTANT_CASEID_KEY] = case_id_key

    if resource_key is not None:
        parameters[constants.PARAMETER_CONSTANT_RESOURCE_KEY] = resource_key

    if group_key is not None:
        parameters[constants.PARAMETER_CONSTANT_GROUP_KEY] = group_key

    for k, v in kwargs.items():
        parameters[k] = v

    return parameters


@deprecation.deprecated(
    deprecated_in="2.3.0",
    removed_in="3.0.0",
    details="This method will be removed in a future release. Please use the method-specific arguments.",
)
def set_classifier(
    log,
    classifier,
    classifier_attribute=constants.DEFAULT_CLASSIFIER_ATTRIBUTE,
):
    """
    Sets the specified classifier on an existing event log.

    :param log: Log object.
    :param classifier: Classifier to set. It can be:
                       - A list of event attributes.
                       - A single event attribute.
                       - A classifier stored in the "classifiers" of the log object.
    :param classifier_attribute: The attribute of the event that will store the concatenation of the attribute values for the given classifier.
    :return: The updated log object as an EventLog or Pandas DataFrame.
    :rtype: Union[EventLog, pd.DataFrame]
    """
    __event_log_deprecation_warning(log)

    if type(classifier) is list:
        pass
    elif type(classifier) is str:
        if type(log) is EventLog and classifier in log.classifiers:
            classifier = log.classifiers[classifier]
        else:
            classifier = [classifier]

    if type(log) is EventLog:
        for trace in log:
            for event in trace:
                event[classifier_attribute] = "+".join(
                    [str(event[x]) for x in classifier]
                )
        log.properties[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] = (
            classifier_attribute
        )
        log.properties[constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY] = (
            classifier_attribute
        )
    elif pandas_utils.check_is_pandas_dataframe(log):
        log[classifier_attribute] = log[classifier[0]]
        for i in range(1, len(classifier)):
            log[classifier_attribute] = (
                log[classifier_attribute] + "+" + log[classifier[i]]
            )
        log.attrs[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] = (
            classifier_attribute
        )
        log.attrs[constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY] = (
            classifier_attribute
        )
    else:
        raise Exception(
            "Setting classifier is not defined for this class of objects"
        )

    return log


def parse_event_log_string(
    traces: Collection[str],
    sep: str = ",",
    activity_key: str = xes_constants.DEFAULT_NAME_KEY,
    timestamp_key: str = xes_constants.DEFAULT_TIMESTAMP_KEY,
    case_id_key: str = constants.CASE_CONCEPT_NAME,
    return_legacy_log_object: bool = constants.DEFAULT_READ_XES_LEGACY_OBJECT,
) -> Union[EventLog, pd.DataFrame]:
    """
    Parses a collection of traces expressed as strings (e.g., ["A,B,C,D", "A,C,B,D", "A,D"]) into a log object.

    :param traces: Collection of traces expressed as strings.
    :param sep: Separator used to split the activities in a string trace.
    :param activity_key: The attribute to be used as activity.
    :param timestamp_key: The attribute to be used as timestamp.
    :param case_id_key: The attribute to be used as case identifier.
    :param return_legacy_log_object: If True, returns a legacy log object (EventLog). If False, returns a Pandas DataFrame. Default is False.
    :return: A log object, either as a legacy EventLog or a Pandas DataFrame.
    :rtype: Union[EventLog, pd.DataFrame]

    .. code-block:: python3

        import pm4py

        dataframe = pm4py.parse_event_log_string(["A,B,C,D", "A,C,B,D", "A,D"])
    """
    cases = []
    activitiess = []
    timestamps = []
    this_timest = 10000000

    for index, trace in enumerate(traces):
        activities = trace.split(sep)
        for act in activities:
            cases.append(str(index))
            activitiess.append(act)
            timestamps.append(
                strpfromiso.fix_naivety(
                    datetime.datetime.fromtimestamp(this_timest)
                )
            )
            this_timest += 1

    dataframe = pandas_utils.instantiate_dataframe(
        {
            case_id_key: cases,
            activity_key: activitiess,
            timestamp_key: timestamps,
        }
    )

    if return_legacy_log_object:
        import pm4py

        return pm4py.convert_to_event_log(dataframe, case_id_key=case_id_key)

    return dataframe


def project_on_event_attribute(
    log: Union[EventLog, pd.DataFrame],
    attribute_key=xes_constants.DEFAULT_NAME_KEY,
    case_id_key=None,
) -> List[List[str]]:
    """
    Projects the event log onto a specified event attribute. The result is a list containing a list for each case, where each case is represented as a list of values for the specified attribute.

    **Example**:

    ```python
    pm4py.project_on_event_attribute(log, "concept:name")
    ```

    **Output**:

    ```python
    [
        ['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request', 'examine thoroughly', 'check ticket', 'decide', 'pay compensation'],
        ['register request', 'check ticket', 'examine casually', 'decide', 'pay compensation'],
        ['register request', 'examine thoroughly', 'check ticket', 'decide', 'reject request'],
        ['register request', 'examine casually', 'check ticket', 'decide', 'pay compensation'],
        ['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request', 'check ticket', 'examine casually', 'decide', 'reinitiate request', 'examine casually', 'check ticket', 'decide', 'reject request'],
        ['register request', 'check ticket', 'examine thoroughly', 'decide', 'reject request']
    ]
    ```

    :param log: Event log or Pandas DataFrame.
    :param attribute_key: The attribute to be used for projection.
    :param case_id_key: (Optional) The attribute to be used as case identifier.
    :return: A list of lists containing the projected attribute values for each case.
    :rtype: List[List[str]]

    .. code-block:: python3

        import pm4py

        list_list_activities = pm4py.project_on_event_attribute(dataframe, 'concept:name')
    """
    __event_log_deprecation_warning(log)

    output = []
    if check_is_pandas_dataframe(log):
        if is_polars_lazyframe(log):
            check_pandas_dataframe_columns(
                log, activity_key=attribute_key, case_id_key=case_id_key
            )

            import polars as pl  # type: ignore[import-untyped]

            lf = log
            schema = lf.collect_schema()
            column_names = schema.names()

            case_column = (
                case_id_key if case_id_key is not None else constants.CASE_CONCEPT_NAME
            )
            if case_column not in column_names:
                raise Exception(
                    f"The specified case identifier column '{case_column}' is not present in the Polars LazyFrame."
                )

            activity_column = (
                attribute_key if attribute_key is not None else xes_constants.DEFAULT_NAME_KEY
            )
            has_activity_column = (
                activity_column is not None and activity_column in column_names
            )

            index_column = constants.DEFAULT_INDEX_KEY
            if index_column not in column_names:
                index_column = "__pm4py_row_index__"
                lf = lf.with_row_count(index_column)
                schema = lf.collect_schema()
                column_names = schema.names()

            sort_columns = [case_column]
            if xes_constants.DEFAULT_TIMESTAMP_KEY in column_names:
                sort_columns.append(xes_constants.DEFAULT_TIMESTAMP_KEY)
            if index_column in column_names:
                sort_columns.append(index_column)

            lf_sorted = lf.sort(sort_columns)

            if has_activity_column:
                lf_selected = lf_sorted.select(
                    pl.col(case_column).alias("__pm_case__"),
                    pl.col(activity_column).alias("__pm_attr__"),
                )
            else:
                lf_selected = lf_sorted.select(
                    pl.col(case_column).alias("__pm_case__"),
                    pl.lit(None).alias("__pm_attr__"),
                )

            aggregated = (
                lf_selected.group_by("__pm_case__", maintain_order=True)
                .agg(pl.col("__pm_attr__").alias("__pm_attr__"))
                .collect()
            )

            output.extend(aggregated["__pm_attr__"].to_list())
        else:
            check_pandas_dataframe_columns(log)
            from pm4py.streaming.conversion import from_pandas

            parameters = {from_pandas.Parameters.ACTIVITY_KEY: attribute_key}
            if case_id_key is not None:
                parameters[from_pandas.Parameters.CASE_ID_KEY] = case_id_key
            it = from_pandas.apply(log, parameters=parameters)
            for trace in it:
                output.append(
                    [
                        (
                            x[xes_constants.DEFAULT_NAME_KEY]
                            if xes_constants.DEFAULT_NAME_KEY is not None
                            else None
                        )
                        for x in trace
                    ]
                )
    else:
        for trace in log:
            output.append(
                [
                    x[attribute_key] if attribute_key is not None else None
                    for x in trace
                ]
            )
    return output


def sample_cases(
    log: Union[EventLog, pd.DataFrame],
    num_cases: int,
    case_id_key: str = "case:concept:name",
) -> Union[EventLog, pd.DataFrame]:
    """
    Randomly samples a given number of cases from the event log.

    :param log: Event log or Pandas DataFrame.
    :param num_cases: Number of cases to sample.
    :param case_id_key: Attribute to be used as case identifier.
    :return: A sampled log object, either as an EventLog or a Pandas DataFrame.
    :rtype: Union[EventLog, pd.DataFrame]

    .. code-block:: python3

        import pm4py

        sampled_dataframe = pm4py.sample_cases(dataframe, 10, case_id_key='case:concept:name')
    """
    __event_log_deprecation_warning(log)

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, case_id_key=case_id_key)

    properties = get_properties(log, case_id_key=case_id_key)

    if isinstance(log, EventLog):
        from pm4py.objects.log.util import sampling

        return sampling.sample(log, num_cases)
    elif check_is_pandas_dataframe(log):
        from pm4py.objects.log.util import dataframe_utils

        properties["max_no_cases"] = num_cases
        return dataframe_utils.sample_dataframe(log, parameters=properties)


def sample_events(
    log: Union[EventStream, OCEL], num_events: int
) -> Union[EventStream, OCEL, pd.DataFrame]:
    """
    Randomly samples a given number of events from the event log.

    :param log: Event stream, OCEL, or Pandas DataFrame.
    :param num_events: Number of events to sample.
    :return: A sampled log object, either as an EventStream, OCEL, or Pandas DataFrame.
    :rtype: Union[EventStream, OCEL, pd.DataFrame]

    .. code-block:: python3

        import pm4py

        sampled_dataframe = pm4py.sample_events(dataframe, 100)
    """
    __event_log_deprecation_warning(log)

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)

    if isinstance(log, EventLog):
        from pm4py.objects.log.util import sampling

        return sampling.sample_log(log, num_events)
    elif isinstance(log, EventStream):
        from pm4py.objects.log.util import sampling

        return sampling.sample_stream(log, num_events)
    elif isinstance(log, OCEL):
        from pm4py.objects.ocel.util import sampling

        return sampling.sample_ocel_events(
            log, parameters={"num_entities": num_events}
        )
    elif check_is_pandas_dataframe(log):
        return log.sample(n=num_events)


def __event_log_deprecation_warning(log):
    if constants.SHOW_EVENT_LOG_DEPRECATION and not hasattr(
        log, "deprecation_warning_shown"
    ):
        if constants.SHOW_INTERNAL_WARNINGS:
            if isinstance(log, EventLog) or isinstance(log, Trace):
                warnings.warn(
                    "The EventLog class has been deprecated and will be removed in a future release."
                )
                log.deprecation_warning_shown = True
            elif isinstance(log, Trace):
                warnings.warn(
                    "The Trace class has been deprecated and will be removed in a future release."
                )
                log.deprecation_warning_shown = True
            elif isinstance(log, EventStream):
                warnings.warn(
                    "The EventStream class has been deprecated and will be removed in a future release."
                )
                log.deprecation_warning_shown = True


def __rustxes_usage_warning():
    if Shared.RUSTXES_WARNING_SHOWN is False:
        warnings.warn("In the current version, the import/export operation uses `rustxes` by default for importing/exporting files faster. Please uninstall `rustxes` to revert the behavior.")
        Shared.RUSTXES_WARNING_SHOWN = True

def __rustxes_non_usage_warning():
    if Shared.RUSTXES_WARNING_SHOWN is False:
        warnings.warn("Install the optional requirement `rustxes` to import/export files faster.")
        Shared.RUSTXES_WARNING_SHOWN = True


def __event_log_filtering_level_warning():
    if Shared.FILTERING_LEVEL_WARNING_SHOWN is False:
        warnings.warn("The filtering operates at the case level (i.e., cases that contain at least one event satisfying the filter are retained in their entirety).")
        Shared.FILTERING_LEVEL_WARNING_SHOWN = True



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.process_tree.obj import ProcessTree, Operator as PTOperator
from pm4py.objects.powl.obj import (
    POWL,
    StrictPartialOrder,
    OperatorPOWL,
    Transition,
    SilentTransition,
)
from typing import Optional, Dict, Any


def apply_recursive(tree: ProcessTree, rec_depth=0) -> POWL:
    """
    Internal method
    """
    nodes = []

    for c in tree.children:
        nodes.append(apply_recursive(c, rec_depth + 1))

    if tree.operator is None:
        if tree.label is not None:
            powl = Transition(label=tree.label)
        else:
            powl = SilentTransition()
    elif tree.operator == PTOperator.OR:
        raise Exception(
            "conversion of process trees containing OR nodes is not supported!"
        )
    elif tree.operator == PTOperator.XOR:
        powl = OperatorPOWL(PTOperator.XOR, nodes)
    elif tree.operator == PTOperator.LOOP:
        powl = OperatorPOWL(PTOperator.LOOP, nodes)
    else:
        powl = StrictPartialOrder(nodes=nodes)

    if tree.operator == PTOperator.SEQUENCE:
        for i in range(len(nodes) - 1):
            powl.order.add_edge(nodes[i], nodes[i + 1])

    return powl


def apply(
    tree: ProcessTree, parameters: Optional[Dict[Any, Any]] = None
) -> POWL:
    """
    Converts a process tree model to a POWL model

    Parameters
    ---------------
    tree
        Process tree

    Returns
    ---------------
    powl_model
        POWL model
    """
    if parameters is None:
        parameters = {}

    return apply_recursive(tree)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys

import numpy as np

from pm4py.objects.random_variables.basic_structure import (
    BasicStructureRandomVariable,
)


class Exponential(BasicStructureRandomVariable):
    """
    Describes a normal variable
    """

    def __init__(self, loc=1, scale=1):
        """
        Constructor

        Parameters
        -----------
        loc
            Loc of the distribution (see docs.scipy.org/doc/scipy/reference/generated/scipy.stats.expon.html)
        scale
            Scale of the distribution
        """
        self.loc = loc
        self.scale = scale
        self.priority = 0
        BasicStructureRandomVariable.__init__(self)

    def read_from_string(self, distribution_parameters):
        """
        Initialize distribution parameters from string

        Parameters
        -----------
        distribution_parameters
            Current distribution parameters as exported on the Petri net
        """
        self.loc = 0
        self.scale = 1.0 / float(distribution_parameters)

    def get_distribution_type(self):
        """
        Get current distribution type

        Returns
        -----------
        distribution_type
            String representing the distribution type
        """
        return "EXPONENTIAL"

    def get_distribution_parameters(self):
        """
        Get a string representing distribution parameters

        Returns
        -----------
        distribution_parameters
            String representing distribution parameters
        """
        if self.scale > 0:
            return str(1.0 / float(self.scale))
        return "UNDEFINED"

    def calculate_loglikelihood(self, values):
        """
        Calculate log likelihood

        Parameters
        ------------
        values
            Empirical values to work on

        Returns
        ------------
        likelihood
            Log likelihood that the values follows the distribution
        """
        from scipy.stats import expon

        if len(values) > 1:
            somma = 0
            for value in values:
                somma = somma + np.log(expon.pdf(value, self.loc, self.scale))
            return somma
        return -sys.float_info.max

    def calculate_parameters(self, values):
        """
        Calculate parameters of the current distribution

        Parameters
        -----------
        values
            Empirical values to work on
        """
        from scipy.stats import expon

        if len(values) > 1:
            self.loc, self.scale = expon.fit(values, floc=0)

    def get_value(self):
        """
        Get a random value following the distribution

        Returns
        -----------
        value
            Value obtained following the distribution
        """
        from scipy.stats import expon

        return expon.rvs(self.loc, self.scale)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.dfg.utils import dfg_utils
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils import petri_utils as pn_util
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"


PARAM_KEY_START_ACTIVITIES = Parameters.START_ACTIVITIES
PARAM_KEY_END_ACTIVITIES = Parameters.END_ACTIVITIES


def apply(dfg, parameters=None):
    """
    Applies the DFG mining on a given object (if it is a Pandas dataframe or a log, the DFG is calculated)

    Parameters
    -------------
    dfg
        Object (DFG) (if it is a Pandas dataframe or a log, the DFG is calculated)
    parameters
        Parameters
    """
    if parameters is None:
        parameters = {}

    dfg = dfg
    start_activities = exec_utils.get_param_value(
        Parameters.START_ACTIVITIES,
        parameters,
        dfg_utils.infer_start_activities(dfg),
    )
    end_activities = exec_utils.get_param_value(
        Parameters.END_ACTIVITIES,
        parameters,
        dfg_utils.infer_end_activities(dfg),
    )
    activities = dfg_utils.get_activities_from_dfg(dfg)

    net = PetriNet("")
    im = Marking()
    fm = Marking()

    source = PetriNet.Place("source")
    net.places.add(source)
    im[source] = 1
    sink = PetriNet.Place("sink")
    net.places.add(sink)
    fm[sink] = 1

    places_corr = {}
    index = 0

    for act in activities:
        places_corr[act] = PetriNet.Place(act)
        net.places.add(places_corr[act])

    for act in start_activities:
        if act in places_corr:
            index = index + 1
            trans = PetriNet.Transition(act + "_" + str(index), act)
            net.transitions.add(trans)
            pn_util.add_arc_from_to(source, trans, net)
            pn_util.add_arc_from_to(trans, places_corr[act], net)

    for act in end_activities:
        if act in places_corr:
            index = index + 1
            inv_trans = PetriNet.Transition(act + "_" + str(index), None)
            net.transitions.add(inv_trans)
            pn_util.add_arc_from_to(places_corr[act], inv_trans, net)
            pn_util.add_arc_from_to(inv_trans, sink, net)

    for el in dfg.keys():
        act1 = el[0]
        act2 = el[1]

        index = index + 1
        trans = PetriNet.Transition(act2 + "_" + str(index), act2)
        net.transitions.add(trans)

        pn_util.add_arc_from_to(places_corr[act1], trans, net)
        pn_util.add_arc_from_to(trans, places_corr[act2], net)

    return net, im, fm



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
import numpy as np
from pm4py.algo.clustering.trace_attribute_driven.util import filter_subsets
from scipy.spatial.distance import pdist
from collections import Counter
from pm4py.util import exec_utils, pandas_utils
from enum import Enum
from pm4py.util import constants


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    SINGLE = "single"
    BINARIZE = "binarize"
    POSITIVE = "positive"
    LOWER_PERCENT = "lower_percent"


def occu_suc(dfg, filter_percent):
    """

    :param dfg: a counter containing all the direct succession relationship with frequency
    :param filter_percent: clarify the percentage of direct succession one wants to preserve
    :return: dataframe of direct succession relationship with frequency
    """

    df = pandas_utils.instantiate_dataframe_from_dict(
        dict(dfg), orient="index", columns=["freq"]
    )
    df = df.sort_values(axis=0, by=["freq"], ascending=False)
    df = df.reset_index().rename(columns={"index": "suc"})
    # delete duplicated successions
    df = df.drop_duplicates("suc", keep="first")
    # delete self succession
    # filter out direct succession by percentage
    filter = list(range(0, round(filter_percent * len(df))))
    df = pandas_utils.insert_index(df)
    df = df[df[constants.DEFAULT_INDEX_KEY].isin(filter)].reset_index(
        drop=True
    )
    return df


def occu_var_suc(var_list, parameters=None):
    """
    return dataframe that shows the frequency of each element(direct succession) in each variant list
    :param var_list:
    :param parameters: binarize states if user wants to binarize the frequency, default is binarized
    :return:
    """
    if parameters is None:
        parameters = {}

    binarize = exec_utils.get_param_value(
        Parameters.BINARIZE, parameters, True
    )

    comb_list = [
        var_list[i] + constants.DEFAULT_VARIANT_SEP + var_list[i + 1]
        for i in range(len(var_list) - 1)
    ]
    result = Counter(comb_list)  # count number of occurrence of each element
    df = pandas_utils.instantiate_dataframe_from_dict(
        dict(result), orient="index", columns=["freq"]
    )
    df = df.reset_index().rename(columns={"index": "direct_suc"})
    if binarize:
        # Binarize succession frequency (optional)
        df.loc[df.freq > 1, "freq"] = 1
        return df
    else:
        return df


def suc_sim(
    var_list_1, var_list_2, log1, log2, freq_thres, num, parameters=None
):
    """

    this function compare the activity similarity between two sublogs via the two lists of variants.
    :param var_list_1: lists of variants in sublog 1
    :param var_list_2: lists of variants in sublog 2
    :param freq_thres: same as sublog2df()
    :param log1: input sublog1 of sublog2df(), which must correspond to var_list_1
    :param log2: input sublog2 of sublog2df(), which must correspond to var_list_2
    :return: the distance matrix between 2 sublogs in which each element is the distance between two variants.
    """

    if parameters is None:
        parameters = {}

    single = exec_utils.get_param_value(Parameters.SINGLE, parameters, False)

    if len(var_list_1) >= len(var_list_2):
        max_len = len(var_list_1)
        min_len = len(var_list_2)
        max_var = var_list_1
        min_var = var_list_2
        var_count_max = filter_subsets.sublog2df(log1, freq_thres, num)[
            "count"
        ]
        var_count_min = filter_subsets.sublog2df(log2, freq_thres, num)[
            "count"
        ]
    else:
        max_len = len(var_list_2)
        min_len = len(var_list_1)
        max_var = var_list_2
        min_var = var_list_1
        var_count_max = filter_subsets.sublog2df(log2, freq_thres, num)[
            "count"
        ]
        var_count_min = filter_subsets.sublog2df(log1, freq_thres, num)[
            "count"
        ]

    dist_matrix = np.zeros((max_len, min_len))
    max_per_var = np.zeros(max_len)
    max_freq = np.zeros(max_len)
    col_sum = np.zeros(max_len)

    if var_list_1 == var_list_2:
        print("Please give different variant lists!")
    else:
        for i in range(max_len):
            dist_vec = np.zeros(min_len)
            df_1 = occu_var_suc(max_var[i], parameters={"binarize": False})
            for j in range(min_len):
                df_2 = occu_var_suc(min_var[j], parameters={"binarize": False})
                df = pandas_utils.merge(
                    df_1, df_2, how="outer", on="direct_suc"
                ).fillna(0)
                # cosine similarity is used to calculate trace similarity
                dist_vec[j] = pdist(
                    np.array([df["freq_x"].values, df["freq_y"].values]),
                    "cosine",
                )[0]

                dist_matrix[i][j] = dist_vec[j]
                if single:
                    if abs(dist_vec[j]) <= 1e-8:
                        max_freq[i] = (
                            var_count_max.iloc[i] * var_count_min.iloc[j]
                        )
                        max_per_var[i] = dist_vec[j] * max_freq[i]

                        break
                    elif j == (min_len - 1):
                        max_loc_col = np.argmin(
                            dist_vec
                        )  # location of max value
                        max_freq[i] = (
                            var_count_max.iloc[i]
                            * var_count_min.iloc[max_loc_col]
                        )
                        max_per_var[i] = dist_vec[max_loc_col] * max_freq[i]
                else:
                    col_sum[i] += (
                        dist_vec[j]
                        * var_count_max.iloc[i]
                        * var_count_min.iloc[j]
                    )

    if single:
        # single linkage
        dist = np.sum(max_per_var) / np.sum(max_freq)
    else:
        vmax_vec = (var_count_max.values).reshape(-1, 1)
        vmin_vec = (var_count_min.values).reshape(1, -1)
        vec_sum = np.sum(np.dot(vmax_vec, vmin_vec))
        dist = np.sum(col_sum) / vec_sum

    return dist


def suc_sim_dual(
    var_list_1, var_list_2, log1, log2, freq_thres, num, parameters=None
):
    """

    this function compare the activity similarity between two sublogs via the two lists of variants.
    :param var_list_1: lists of variants in sublog 1
    :param var_list_2: lists of variants in sublog 2
    :param freq_thres: same as sublog2df()
    :param log1: input sublog1 of sublog2df(), which must correspond to var_list_1
    :param log2: input sublog2 of sublog2df(), which must correspond to var_list_2
    :return: the distance matrix between 2 sublogs in which each element is the distance between two variants.
    """

    if parameters is None:
        parameters = {}

    single = exec_utils.get_param_value(Parameters.SINGLE, parameters, False)

    if len(var_list_1) >= len(var_list_2):
        max_len = len(var_list_1)
        min_len = len(var_list_2)
        max_var = var_list_1
        min_var = var_list_2
        var_count_max = filter_subsets.sublog2df(log1, freq_thres, num)[
            "count"
        ]
        var_count_min = filter_subsets.sublog2df(log2, freq_thres, num)[
            "count"
        ]
    else:
        max_len = len(var_list_2)
        min_len = len(var_list_1)
        max_var = var_list_2
        min_var = var_list_1
        var_count_max = filter_subsets.sublog2df(log2, freq_thres, num)[
            "count"
        ]
        var_count_min = filter_subsets.sublog2df(log1, freq_thres, num)[
            "count"
        ]

    dist_matrix = np.zeros((max_len, min_len))
    max_per_var = np.zeros(max_len)
    max_freq = np.zeros(max_len)
    min_freq = np.zeros(min_len)
    min_per_var = np.zeros(min_len)
    col_sum = np.zeros(max_len)
    index_rec = set(list(range(min_len)))

    if var_list_1 == var_list_2:
        print("Please give different variant lists!")
    else:
        for i in range(max_len):
            dist_vec = np.zeros(min_len)
            df_1 = occu_var_suc(max_var[i], parameters={"binarize": False})
            for j in range(min_len):
                df_2 = occu_var_suc(min_var[j], parameters={"binarize": False})
                df = pandas_utils.merge(
                    df_1, df_2, how="outer", on="direct_suc"
                ).fillna(0)
                # cosine similarity is used to calculate trace similarity
                dist_vec[j] = pdist(
                    np.array([df["freq_x"].values, df["freq_y"].values]),
                    "cosine",
                )[0]
                dist_matrix[i][j] = dist_vec[j]

                if j == (min_len - 1):
                    max_loc_col = np.argmin(dist_vec)
                    if abs(dist_vec[max_loc_col]) <= 1e-8:
                        index_rec.discard(max_loc_col)
                        max_freq[i] = (
                            var_count_max.iloc[i]
                            * var_count_min.iloc[max_loc_col]
                            * 2
                        )
                        max_per_var[i] = (
                            dist_vec[max_loc_col] * max_freq[i] * 2
                        )
                    else:
                        max_freq[i] = (
                            var_count_max.iloc[i]
                            * var_count_min.iloc[max_loc_col]
                        )
                        max_per_var[i] = dist_vec[max_loc_col] * max_freq[i]

        if len(index_rec) != 0:
            for i in list(index_rec):
                min_loc_row = np.argmin(dist_matrix[:, i])
                min_freq[i] = (
                    var_count_max.iloc[min_loc_row] * var_count_min.iloc[i]
                )
                min_per_var[i] = dist_matrix[min_loc_row, i] * min_freq[i]

    if single:
        # single linkage
        dist = np.sum(max_per_var) / np.sum(max_freq)
    else:
        vmax_vec = (var_count_max.values).reshape(-1, 1)
        vmin_vec = (var_count_min.values).reshape(1, -1)
        vec_sum = np.sum(np.dot(vmax_vec, vmin_vec))
        dist = np.sum(col_sum) / vec_sum

    return dist


def suc_sim_percent(log1, log2, percent_1, percent_2):
    """

    this function compare the activity similarity between two sublogs via the two lists of variants.
    :param var_list_1: lists of variants in sublog 1
    :param var_list_2: lists of variants in sublog 2
    :param freq_thres: same as sublog2df()
    :param log1: input sublog1 of sublog2df(), which must correspond to var_list_1
    :param log2: input sublog2 of sublog2df(), which must correspond to var_list_2
    :return: the distance matrix between 2 sublogs in which each element is the distance between two variants.
    """

    (dataframe_1, var_list_1) = filter_subsets.sublog_percent(log1, percent_1)
    (dataframe_2, var_list_2) = filter_subsets.sublog_percent(log2, percent_2)

    if len(var_list_1) >= len(var_list_2):
        max_len = len(var_list_1)
        min_len = len(var_list_2)
        max_var = var_list_1
        min_var = var_list_2
        var_count_max = dataframe_1["count"]
        var_count_min = dataframe_2["count"]
    else:
        max_len = len(var_list_2)
        min_len = len(var_list_1)
        max_var = var_list_2
        min_var = var_list_1
        var_count_max = dataframe_2["count"]
        var_count_min = dataframe_1["count"]

    dist_matrix = np.zeros((max_len, min_len))
    max_per_var = np.zeros(max_len)
    max_freq = np.zeros(max_len)
    min_freq = np.zeros(min_len)
    min_per_var = np.zeros(min_len)
    col_sum = np.zeros(max_len)
    index_rec = set(list(range(min_len)))

    if var_list_1 == var_list_2:
        dist = 0
    else:
        for i in range(max_len):
            dist_vec = np.zeros(min_len)
            df_1 = occu_var_suc(max_var[i], parameters={"binarize": False})
            for j in range(min_len):
                df_2 = occu_var_suc(min_var[j], parameters={"binarize": False})
                df = pandas_utils.merge(
                    df_1, df_2, how="outer", on="direct_suc"
                ).fillna(0)
                # cosine similarity is used to calculate trace similarity
                dist_vec[j] = pdist(
                    np.array([df["freq_x"].values, df["freq_y"].values]),
                    "cosine",
                )[0]
                if np.isnan(dist_vec[j]):
                    dist_vec[j] = 1
                dist_matrix[i][j] = dist_vec[j]
                if j == (min_len - 1):
                    max_loc_col = np.argmin(dist_vec)
                    if abs(dist_vec[max_loc_col]) <= 1e-8:
                        index_rec.discard(max_loc_col)
                        max_freq[i] = (
                            var_count_max.iloc[i]
                            * var_count_min.iloc[max_loc_col]
                            * 2
                        )
                        max_per_var[i] = (
                            dist_vec[max_loc_col] * max_freq[i] * 2
                        )
                    else:
                        max_freq[i] = (
                            var_count_max.iloc[i]
                            * var_count_min.iloc[max_loc_col]
                        )
                        max_per_var[i] = dist_vec[max_loc_col] * max_freq[i]

        if len(index_rec) != 0:
            for i in list(index_rec):
                min_loc_row = np.argmin(dist_matrix[:, i])
                min_freq[i] = (
                    var_count_max.iloc[min_loc_row] * var_count_min.iloc[i]
                )
                min_per_var[i] = dist_matrix[min_loc_row, i] * min_freq[i]

        # single linkage
        dist = (np.sum(max_per_var) + np.sum(min_per_var)) / (
            np.sum(max_freq) + np.sum(min_freq)
        )

    return dist


def suc_sim_percent_avg(log1, log2, percent_1, percent_2):
    """

    this function compare the activity similarity between two sublogs via the two lists of variants.
    :param var_list_1: lists of variants in sublog 1
    :param var_list_2: lists of variants in sublog 2
    :param freq_thres: same as sublog2df()
    :param log1: input sublog1 of sublog2df(), which must correspond to var_list_1
    :param log2: input sublog2 of sublog2df(), which must correspond to var_list_2
    :return: the distance matrix between 2 sublogs in which each element is the distance between two variants.
    """

    (dataframe_1, var_list_1) = filter_subsets.sublog_percent(log1, percent_1)
    (dataframe_2, var_list_2) = filter_subsets.sublog_percent(log2, percent_2)

    if len(var_list_1) >= len(var_list_2):
        max_len = len(var_list_1)
        min_len = len(var_list_2)
        max_var = var_list_1
        min_var = var_list_2
        var_count_max = dataframe_1["count"]
        var_count_min = dataframe_2["count"]
    else:
        max_len = len(var_list_2)
        min_len = len(var_list_1)
        max_var = var_list_2
        min_var = var_list_1
        var_count_max = dataframe_2["count"]
        var_count_min = dataframe_1["count"]

    dist_matrix = np.zeros((max_len, min_len))
    col_sum = np.zeros(max_len)

    for i in range(max_len):
        dist_vec = np.zeros(min_len)
        df_1 = occu_var_suc(max_var[i], parameters={"binarize": False})
        for j in range(min_len):
            df_2 = occu_var_suc(min_var[j], parameters={"binarize": False})
            df = pandas_utils.merge(
                df_1, df_2, how="outer", on="direct_suc"
            ).fillna(0)
            # cosine similarity is used to calculate trace similarity
            dist_vec[j] = pdist(
                np.array([df["freq_x"].values, df["freq_y"].values]), "cosine"
            )[0]
            if np.isnan(dist_vec[j]):
                dist_vec[j] = 1
            dist_matrix[i][j] = dist_vec[j]
            col_sum[i] += (
                dist_vec[j] * var_count_max.iloc[i] * var_count_min.iloc[j]
            )

    # single linkage
    vmax_vec = (var_count_max.values).reshape(-1, 1)
    vmin_vec = (var_count_min.values).reshape(1, -1)
    vec_sum = np.sum(np.dot(vmax_vec, vmin_vec))
    dist = np.sum(col_sum) / vec_sum

    return dist



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.util import exec_utils, pandas_utils, constants
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ATTRIBUTE_KEY,
    PARAMETER_CONSTANT_CASEID_KEY,
    PARAMETER_CONSTANT_RESOURCE_KEY,
    PARAMETER_CONSTANT_TIMESTAMP_KEY,
)
from pm4py.util.xes_constants import (
    DEFAULT_NAME_KEY,
    DEFAULT_RESOURCE_KEY,
    DEFAULT_TIMESTAMP_KEY,
)
from copy import copy
from typing import Optional, Dict, Any, Union, List
import pandas as pd


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = PARAMETER_CONSTANT_ATTRIBUTE_KEY
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    RESOURCE_KEY = PARAMETER_CONSTANT_RESOURCE_KEY
    POSITIVE = "positive"
    ENABLE_TIMESTAMP = "enable_timestamp"
    TIMESTAMP_DIFF_BOUNDARIES = "timestamp_diff_boundaries"


POSITIVE = Parameters.POSITIVE
ENABLE_TIMESTAMP = Parameters.ENABLE_TIMESTAMP
TIMESTAMP_DIFF_BOUNDARIES = Parameters.TIMESTAMP_DIFF_BOUNDARIES


def eventually_follows(
    df0: pd.DataFrame,
    attribute_values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pd.DataFrame:
    """
    Applies the eventually follows rule

    Parameters
    ------------
    df0
        Dataframe
    attribute_values
        A list of attribute_values attribute_values[n] follows attribute_values[n-1] follows ... follows attribute_values[0]

    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
        - If True, returns all the cases containing all attribute_values and in which attribute_values[i] was eventually followed by attribute_values[i + 1]
        - If False, returns all the cases not containing all attribute_values, or in which an instance of attribute_values[i] was not eventually
        followed by an instance of attribute_values[i + 1]

    Returns
    ------------
    filtered_df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    timestamp_diff_boundaries = exec_utils.get_param_value(
        Parameters.TIMESTAMP_DIFF_BOUNDARIES, parameters, []
    )
    enable_timestamp = exec_utils.get_param_value(
        Parameters.ENABLE_TIMESTAMP,
        parameters,
        len(timestamp_diff_boundaries) > 0,
    )

    colset = [case_id_glue, attribute_key]
    if enable_timestamp:
        colset.append(timestamp_key)

    df = df0.copy()
    df = df[colset]
    df = pandas_utils.insert_index(df)

    df_a = [
        df[df[attribute_key] == attribute_value].copy()
        for attribute_value in attribute_values
    ]

    df_join = (
        df_a[0].merge(df_a[1], on=case_id_glue, suffixes=("_0", "_1")).dropna()
    )
    df_join[constants.DEFAULT_INDEX_KEY] = df_join[
        constants.DEFAULT_INDEX_KEY + "_0"
    ]
    df_join["@@diffindex0"] = (
        df_join[constants.DEFAULT_INDEX_KEY + "_1"]
        - df_join[constants.DEFAULT_INDEX_KEY + "_0"]
    )
    df_join = df_join[df_join["@@diffindex0"] > 0]

    for i in range(2, len(df_a)):
        df_join = df_join.merge(
            df_a[i], on=case_id_glue, suffixes=("", "_%d" % i)
        ).dropna()
        df_join["@@diffindex%d" % (i - 1)] = (
            df_join[constants.DEFAULT_INDEX_KEY + "_%d" % i]
            - df_join[constants.DEFAULT_INDEX_KEY + "_%d" % (i - 1)]
        )
        df_join = df_join[df_join["@@diffindex%d" % (i - 1)] > 0]

    if enable_timestamp:
        for i in range(1, len(df_a)):
            df_join["@@difftimestamp%d" % (i - 1)] = (
                pandas_utils.get_total_seconds(
                    df_join[timestamp_key + "_%d" % i]
                    - df_join[timestamp_key + "_%d" % (i - 1)]
                )
            )

            if timestamp_diff_boundaries:
                df_join = df_join[
                    df_join["@@difftimestamp%d" % (i - 1)]
                    >= timestamp_diff_boundaries[i - 1][0]
                ]
                df_join = df_join[
                    df_join["@@difftimestamp%d" % (i - 1)]
                    <= timestamp_diff_boundaries[i - 1][1]
                ]

    i1 = df.set_index(case_id_glue).index
    i2 = df_join.set_index(case_id_glue).index
    if positive:
        ret = df0[i1.isin(i2)]
    else:
        ret = df0[~i1.isin(i2)]

    ret.attrs = copy(df0.attrs) if hasattr(df0, "attrs") else {}
    return ret


def A_next_B_next_C(
    df0: pd.DataFrame,
    A: str,
    B: str,
    C: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pd.DataFrame:
    """
    Applies the A net B next C rule

    Parameters
    ------------
    df0
        Dataframe
    A
        A Attribute value
    B
        B Attribute value
    C
        C Attribute value
    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
        - If True, returns all the cases containing A, B and C and in which A was directly followed by B and B was directly followed by C
        - If False, returns all the cases not containing A or B or C, or in which none instance of A was directly
        followed by an instance of B and B was directly followed by C

    Returns
    ------------
    filtered_df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    df = df0.copy()
    df = df[[case_id_glue, attribute_key]]
    df = pandas_utils.insert_index(df)
    df_A = df[df[attribute_key] == A].copy()
    df_B = df[df[attribute_key] == B].copy()
    df_C = df[df[attribute_key] == C].copy()
    df_B["@@conceptname"] = df_B[case_id_glue]
    df_B = df_B.groupby(case_id_glue).last().set_index("@@conceptname")
    df_C["@@conceptname"] = df_C[case_id_glue]
    df_C = df_C.groupby(case_id_glue).last().set_index("@@conceptname")

    df_join = (
        df_A.join(df_B, on=case_id_glue, rsuffix="_2")
        .dropna()
        .join(df_C, on=case_id_glue, rsuffix="_3")
        .dropna()
    )
    df_join["@@diffindex"] = (
        df_join[constants.DEFAULT_INDEX_KEY + "_2"]
        - df_join[constants.DEFAULT_INDEX_KEY]
    )
    df_join["@@diffindex2"] = (
        df_join[constants.DEFAULT_INDEX_KEY + "_3"]
        - df_join[constants.DEFAULT_INDEX_KEY + "_2"]
    )
    df_join = df_join[df_join["@@diffindex"] == 1]
    df_join = df_join[df_join["@@diffindex2"] == 1]

    i1 = df.set_index(case_id_glue).index
    i2 = df_join.set_index(case_id_glue).index

    if positive:
        ret = df0[i1.isin(i2)]
    else:
        ret = df0[~i1.isin(i2)]

    ret.attrs = copy(df0.attrs) if hasattr(df0, "attrs") else {}
    return ret


def four_eyes_principle(
    df0: pd.DataFrame,
    A: str,
    B: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pd.DataFrame:
    """
    Verifies the Four Eyes Principle given A and B

    Parameters
    -------------
    df0
        Dataframe
    A
        A attribute value
    B
        B attribute value
    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
        - if True, then filters all the cases containing A and B which have empty intersection between the set
          of resources doing A and B
        - if False, then filters all the cases containing A and B which have no empty intersection between the set
          of resources doing A and B

    Returns
    --------------
    filtered_df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, DEFAULT_RESOURCE_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    df = df0.copy()
    df = df[[case_id_glue, attribute_key, resource_key]]

    df_A = df[df[attribute_key] == A].copy()
    df_B = df[df[attribute_key] == B].copy()
    df_B["@@conceptname"] = df_B[case_id_glue]
    df_B = df_B.groupby(case_id_glue).last().set_index("@@conceptname")

    df_join = df_A.join(df_B, on=case_id_glue, rsuffix="_2").dropna()
    df_join_pos = df_join[
        df_join[resource_key] == df_join[resource_key + "_2"]
    ]
    df_join_neg = df_join[
        df_join[resource_key] != df_join[resource_key + "_2"]
    ]

    i1 = df.set_index(case_id_glue).index
    i2 = df_join_pos.set_index(case_id_glue).index
    i3 = df_join_neg.set_index(case_id_glue).index

    if positive:
        ret = df0[i1.isin(i3) & ~i1.isin(i2)]
    else:
        ret = df0[i1.isin(i2)]

    ret.attrs = copy(df0.attrs) if hasattr(df0, "attrs") else {}
    return ret


def attr_value_different_persons(
    df0: pd.DataFrame,
    A: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pd.DataFrame:
    """
    Checks whether an attribute value is assumed on events done by different resources

    Parameters
    ------------
    df0
        Dataframe
    A
        A attribute value
    parameters
        Parameters of the algorithm, including the attribute key and the positive parameter:
            - if True, then filters all the cases containing occurrences of A done by different resources
            - if False, then filters all the cases not containing occurrences of A done by different resources

    Returns
    -------------
    filtered_df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, DEFAULT_RESOURCE_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    df = df0.copy()
    df = df[[case_id_glue, attribute_key, resource_key]]

    df_A = df[df[attribute_key] == A].copy()
    df_B = df[df[attribute_key] == A].copy()
    df_B["@@conceptname"] = df_B[case_id_glue]
    df_B = df_B.groupby(case_id_glue).last().set_index("@@conceptname")

    df_join = df_A.join(df_B, on=case_id_glue, rsuffix="_2").dropna()
    df_join_neg = df_join[
        df_join[resource_key] != df_join[resource_key + "_2"]
    ]

    i1 = df.set_index(case_id_glue).index
    i2 = df_join_neg.set_index(case_id_glue).index

    if positive:
        ret = df0[i1.isin(i2)]
    else:
        ret = df0[~i1.isin(i2)]

    ret.attrs = copy(df0.attrs) if hasattr(df0, "attrs") else {}
    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.xes_constants import DEFAULT_NAME_KEY, DEFAULT_TIMESTAMP_KEY
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util import exec_utils
from pm4py.util import constants
from enum import Enum
from typing import Optional, Dict, Any
import polars as pl


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    MAX_NO_POINTS_SAMPLE = "max_no_of_points_to_sample"
    KEEP_ONCE_PER_CASE = "keep_once_per_case"
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def apply(
    lf: pl.LazyFrame,
    activity: str,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Gets the time passed from each preceding activity

    Parameters
    -------------
    lf
        LazyFrame
    activity
        Activity that we are considering
    parameters
        Possible parameters of the algorithm

    Returns
    -------------
    dictio
        Dictionary containing a 'pre' key with the
        list of aggregates times from each preceding activity to the given activity
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )

    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )
    workcalendar = exec_utils.get_param_value(
        Parameters.WORKCALENDAR,
        parameters,
        constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    )

    # Basic DFG computation for preceding activities
    # This is a simplified version - full implementation would need more complex DFG logic
    if start_timestamp_key is None:
        start_timestamp_key = timestamp_key
    
    # Sort by case and timestamp
    sorted_lf = lf.sort([case_id_glue, timestamp_key])
    
    # Create pairs of consecutive activities
    with_next = sorted_lf.with_columns([
        pl.col(activity_key).shift(-1).over(case_id_glue).alias(activity_key + "_next"),
        pl.col(timestamp_key).shift(-1).over(case_id_glue).alias(timestamp_key + "_next"),
    ])
    
    # Filter for target activity as next activity
    target_flows = with_next.filter(
        pl.col(activity_key + "_next") == activity
    )
    
    # Calculate time differences
    if business_hours:
        # For business hours, we need to collect and apply the function
        # This is a simplified version
        flows_df = target_flows.collect()
        # Would need actual business hours calculation here
        flows_df = flows_df.with_columns(
            (pl.col(timestamp_key + "_next") - pl.col(timestamp_key)).dt.total_seconds().alias("flow_time")
        )
    else:
        flows_df = target_flows.with_columns(
            (pl.col(timestamp_key + "_next") - pl.col(timestamp_key)).dt.total_seconds().alias("flow_time")
        ).collect()
    
    # Group by preceding activity and compute statistics
    stats = flows_df.group_by(activity_key).agg([
        pl.col("flow_time").mean().alias("avg_time"),
        pl.col("flow_time").count().alias("frequency")
    ])
    
    pre = []
    sum_perf_pre = 0.0
    sum_acti_pre = 0.0
    
    for row in stats.iter_rows():
        pre_activity = row[0]
        avg_time = row[1]
        frequency = row[2]
        
        pre.append([pre_activity, float(avg_time), int(frequency)])
        sum_perf_pre += float(avg_time) * int(frequency)
        sum_acti_pre += int(frequency)

    if sum_acti_pre > 0:
        sum_perf_pre = sum_perf_pre / sum_acti_pre

    return {"pre": pre, "pre_avg_perf": sum_perf_pre, "pre_total_acti": sum_acti_pre}


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
"""Polars LazyFrame implementation of the disconnected performance spectrum variant."""

from enum import Enum
from typing import Any, Dict, List, Optional, Union

import polars as pl

from pm4py.util import constants, points_subset
from pm4py.util import exec_utils
from pm4py.util import xes_constants as xes


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    PARAMETER_SAMPLE_SIZE = "sample_size"
    SORT_LOG_REQUIRED = "sort_log_required"


def gen_patterns(pattern: List[str], length: int) -> List[str]:
    return ["".join(pattern[i : i + length]) for i in range(len(pattern) - (length - 1))]


def _prepare_dataframe(
    lf: pl.LazyFrame,
    activities: List[str],
    case_id_glue: str,
    activity_key: str,
    timestamp_key: str,
    sort_required: bool,
) -> pl.LazyFrame:
    lf = lf.select(
        pl.col(case_id_glue),
        pl.col(activity_key),
        pl.col(timestamp_key),
    )
    lf = lf.filter(pl.col(activity_key).is_in(activities))
    lf = lf.with_row_count(constants.DEFAULT_EVENT_INDEX_KEY)
    if sort_required:
        lf = lf.sort([case_id_glue, timestamp_key, constants.DEFAULT_EVENT_INDEX_KEY])
    lf = lf.with_columns(pl.col(timestamp_key).dt.timestamp().alias(timestamp_key))
    return lf


def apply(
    dataframe: pl.LazyFrame,
    list_activities: List[str],
    sample_size: int,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Any]:
    """Compute disconnected performance spectrum using a Polars LazyFrame."""

    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    sort_log_required = exec_utils.get_param_value(
        Parameters.SORT_LOG_REQUIRED, parameters, True
    )

    lf = _prepare_dataframe(
        dataframe,
        list_activities,
        case_id_glue,
        activity_key,
        timestamp_key,
        sort_log_required,
    )

    points: List[Dict[str, Any]] = []

    all_patterns = [
        (
            len(list_activities) - i,
            gen_patterns(list_activities, len(list_activities) - i),
        )
        for i in range(len(list_activities) - 1)
    ]

    def key(prefix: str, idx: int) -> str:
        return f"{prefix}_{idx}"

    for length, patterns in all_patterns:
        if length <= 0:
            continue

        pattern_lf = lf
        exprs: List[pl.Expr] = []
        for i in range(length):
            exprs.extend(
                [
                    pl.col(activity_key)
                    .shift(-i)
                    .over(case_id_glue)
                    .alias(key(activity_key, i)),
                    pl.col(timestamp_key)
                    .shift(-i)
                    .over(case_id_glue)
                    .alias(key(timestamp_key, i)),
                    pl.col(constants.DEFAULT_EVENT_INDEX_KEY)
                    .shift(-i)
                    .over(case_id_glue)
                    .alias(key(constants.DEFAULT_EVENT_INDEX_KEY, i)),
                ]
            )

        pattern_lf = pattern_lf.with_columns(exprs)
        pattern_lf = pattern_lf.filter(
            pl.all_horizontal(
                [pl.col(key(activity_key, i)).is_not_null() for i in range(length)]
            )
        )

        pattern_values = ["".join(pat) for pat in patterns]
        if not pattern_values:
            continue

        pattern_lf = pattern_lf.with_columns(
            pl.concat_str([pl.col(key(activity_key, i)) for i in range(length)])
            .alias("pattern")
        )
        pattern_lf = pattern_lf.filter(pl.col("pattern").is_in(pattern_values))

        collected = pattern_lf.collect()
        if collected.height == 0:
            continue

        to_drop_indexes: set[int] = set()

        for row in collected.iter_rows(named=True):
            activities_seq = [row[key(activity_key, i)] for i in range(length)]
            timestamps_seq = [row[key(timestamp_key, i)] / 1_000_000 for i in range(length)]
            case_id = row[case_id_glue]

            points.append(
                {
                    "case_id": case_id,
                    "points": list(zip(activities_seq, timestamps_seq)),
                }
            )

            for i in range(length - 1):
                to_drop_indexes.add(row[key(constants.DEFAULT_EVENT_INDEX_KEY, i)])

        if to_drop_indexes:
            lf = lf.filter(
                ~pl.col(constants.DEFAULT_EVENT_INDEX_KEY).is_in(
                    pl.Series(list(to_drop_indexes))
                )
            )

    points = sorted(
        points,
        key=lambda entry: min(entry["points"], key=lambda evt: evt[1])[1],
    )

    if len(points) > sample_size:
        points = points_subset.pick_chosen_points_list(sample_size, points)

    return points



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.evaluation.precision.variants import etconformance_token
from pm4py.algo.evaluation.precision.variants import align_etconformance
from pm4py.algo.evaluation.precision.variants import automaton_after_align
from pm4py.objects.petri_net.utils.check_soundness import (
    check_easy_soundness_net_in_fin_marking,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.petri_net.obj import PetriNet, Marking
import pandas as pd


class Variants(Enum):
    ETCONFORMANCE_TOKEN = etconformance_token
    ALIGN_ETCONFORMANCE = align_etconformance
    AUTOMATON_AFTER_ALIGN = automaton_after_align


ETCONFORMANCE_TOKEN = Variants.ETCONFORMANCE_TOKEN
ALIGN_ETCONFORMANCE = Variants.ALIGN_ETCONFORMANCE
AUTOMATON_AFTER_ALIGN = Variants.AUTOMATON_AFTER_ALIGN

VERSIONS = {ETCONFORMANCE_TOKEN, ALIGN_ETCONFORMANCE}


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    net: PetriNet,
    marking: Marking,
    final_marking: Marking,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=None,
) -> float:
    """
    Method to apply ET Conformance

    Parameters
    -----------
    log
        Trace log
    net
        Petri net
    marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters of the algorithm, including:
            pm4py.util.constants.PARAMETER_CONSTANT_ACTIVITY_KEY -> Activity key
    variant
        Variant of the algorithm that should be applied:
            - Variants.ETCONFORMANCE_TOKEN
            - Variants.ALIGN_ETCONFORMANCE
    """
    if parameters is None:
        parameters = {}

    # execute the following part of code when the variant is not specified by
    # the user
    if variant is None:
        if not (
            check_easy_soundness_net_in_fin_marking(
                net, marking, final_marking
            )
        ):
            # in the case the net is not a easy sound workflow net, we must
            # apply token-based replay
            variant = ETCONFORMANCE_TOKEN
        else:
            # otherwise, use the align-etconformance approach (safer, in the
            # case the model contains duplicates)
            variant = ALIGN_ETCONFORMANCE

    return exec_utils.get_variant(variant).apply(
        log, net, marking, final_marking, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME, PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, Set
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    MIN_OCCURRENCES = "min_occurrences"


def apply(
    df: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filters the cases where the specified activity occurs at least min_occurrences times.

    Parameters
    -----------
    df
        LazyFrame
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity
            Parameters.MIN_OCCURRENCES -> Minimum occurrences of the activity (default: 2)

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    min_occurrences = exec_utils.get_param_value(
        Parameters.MIN_OCCURRENCES, parameters, 2
    )

    # Count occurrences of each activity per case
    activity_counts = (
        df.group_by([case_id_glue, activity_key])
        .count()
        .filter(pl.col("count") >= min_occurrences)
        .select(case_id_glue)
        .unique()
    )

    # Keep only cases with rework
    ret = df.join(activity_counts, on=case_id_glue, how="inner")

    return ret


def apply_activity_set(
    df: pl.LazyFrame,
    activities: Set[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filters the cases where at least one of the specified activities occurs at least min_occurrences times.

    Parameters
    -----------
    df
        LazyFrame
    activities
        Set of activities to check for rework
    parameters
        Parameters of the algorithm

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    min_occurrences = exec_utils.get_param_value(
        Parameters.MIN_OCCURRENCES, parameters, 2
    )

    # Filter for specified activities and count occurrences
    activity_counts = (
        df.filter(pl.col(activity_key).is_in(list(activities)))
        .group_by([case_id_glue, activity_key])
        .count()
        .filter(pl.col("count") >= min_occurrences)
        .select(case_id_glue)
        .unique()
    )

    # Keep only cases with rework
    ret = df.join(activity_counts, on=case_id_glue, how="inner")

    return ret
